# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log
from zope.interface import implementer

from buildbot import config
from buildbot import util
from buildbot.changes.filter import ChangeFilter
from buildbot.interfaces import ITriggerableScheduler
from buildbot.process import buildstep
from buildbot.process import properties
from buildbot.schedulers import base
from buildbot.util import croniter
from buildbot.util.codebase import AbsoluteSourceStampsMixin


class Timed(AbsoluteSourceStampsMixin, base.BaseScheduler):

    """
    Parent class for timed schedulers.  This takes care of the (surprisingly
    subtle) mechanics of ensuring that each timed actuation runs to completion
    before the service stops.
    """

    compare_attrs = ('reason', 'createAbsoluteSourceStamps', 'onlyIfChanged',
                     'branch', 'fileIsImportant', 'change_filter', 'onlyImportant')
    reason = ''

    class NoBranch:
        pass

    def __init__(self, name, builderNames, reason='',
                 createAbsoluteSourceStamps=False, onlyIfChanged=False,
                 branch=NoBranch, change_filter=None, fileIsImportant=None,
                 onlyImportant=False, **kwargs):
        super().__init__(name, builderNames, **kwargs)

        # tracking for when to start the next build
        self.lastActuated = None

        # A lock to make sure that each actuation occurs without interruption.
        # This lock governs actuateAt, actuateAtTimer, and actuateOk
        self.actuationLock = defer.DeferredLock()
        self.actuateOk = False
        self.actuateAt = None
        self.actuateAtTimer = None

        self.reason = util.bytes2unicode(reason % {'name': name})
        self.branch = branch
        self.change_filter = ChangeFilter.fromSchedulerConstructorArgs(
            change_filter=change_filter)
        self.createAbsoluteSourceStamps = createAbsoluteSourceStamps
        self.onlyIfChanged = onlyIfChanged
        if fileIsImportant and not callable(fileIsImportant):
            config.error(
                "fileIsImportant must be a callable")
        self.fileIsImportant = fileIsImportant
        # If True, only important changes will be added to the buildset.
        self.onlyImportant = onlyImportant
        self._reactor = reactor  # patched by tests

    @defer.inlineCallbacks
    def activate(self):
        yield super().activate()

        if not self.enabled:
            return None

        # no need to lock this
        # nothing else can run before the service is started
        self.actuateOk = True

        # get the scheduler's last_build time (note: only done at startup)
        self.lastActuated = yield self.getState('last_build', None)

        # schedule the next build
        yield self.scheduleNextBuild()

        if self.onlyIfChanged or self.createAbsoluteSourceStamps:
            yield self.startConsumingChanges(fileIsImportant=self.fileIsImportant,
                                             change_filter=self.change_filter,
                                             onlyImportant=self.onlyImportant)
        else:
            yield self.master.db.schedulers.flushChangeClassifications(self.serviceid)
        return None

    @defer.inlineCallbacks
    def deactivate(self):
        yield super().deactivate()

        if not self.enabled:
            return None

        # shut down any pending actuation, and ensure that we wait for any
        # current actuation to complete by acquiring the lock.  This ensures
        # that no build will be scheduled after deactivate is complete.
        def stop_actuating():
            self.actuateOk = False
            self.actuateAt = None
            if self.actuateAtTimer:
                self.actuateAtTimer.cancel()
            self.actuateAtTimer = None
        yield self.actuationLock.run(stop_actuating)
        return None

    # Scheduler methods

    def gotChange(self, change, important):
        # both important and unimportant changes on our branch are recorded, as
        # we will include all such changes in any buildsets we start.  Note
        # that we must check the branch here because it is not included in the
        # change filter.
        if self.branch is not Timed.NoBranch and change.branch != self.branch:
            return defer.succeed(None)  # don't care about this change

        d = self.master.db.schedulers.classifyChanges(
            self.serviceid, {change.number: important})

        if self.createAbsoluteSourceStamps:
            d.addCallback(lambda _: self.recordChange(change))

        return d

    @defer.inlineCallbacks
    def startBuild(self):
        if not self.enabled:
            log.msg(format='ignoring build from %(name)s because scheduler '
                           'has been disabled by the user',
                    name=self.name)
            return

        # use the collected changes to start a build
        scheds = self.master.db.schedulers
        classifications = yield scheds.getChangeClassifications(self.serviceid)

        # if onlyIfChanged is True, then we will skip this build if no important changes have
        # occurred since the last invocation. Note that when the scheduler has just been started
        # there may not be any important changes yet and we should start the build for the
        # current state of the code whatever it is.
        last_only_if_changed = yield self.getState('last_only_if_changed', False)
        if last_only_if_changed and self.onlyIfChanged and not any(classifications.values()):
            log.msg(("{} scheduler <{}>: skipping build " +
                     "- No important changes").format(self.__class__.__name__, self.name))
            return

        if last_only_if_changed != self.onlyIfChanged:
            yield self.setState('last_only_if_changed', self.onlyIfChanged)

        changeids = sorted(classifications.keys())

        if changeids:
            max_changeid = changeids[-1]  # (changeids are sorted)
            yield self.addBuildsetForChanges(reason=self.reason,
                                             changeids=changeids)
            yield scheds.flushChangeClassifications(self.serviceid,
                                                    less_than=max_changeid + 1)
        else:
            # There are no changes, but onlyIfChanged is False, so start
            # a build of the latest revision, whatever that is
            sourcestamps = [dict(codebase=cb) for cb in self.codebases]
            yield self.addBuildsetForSourceStampsWithDefaults(
                reason=self.reason,
                sourcestamps=sourcestamps)

    def getCodebaseDict(self, codebase):
        if self.createAbsoluteSourceStamps:
            return super().getCodebaseDict(codebase)
        return self.codebases[codebase]

    # Timed methods

    def getNextBuildTime(self, lastActuation):
        """
        Called by to calculate the next time to actuate a BuildSet.  Override
        in subclasses.  To trigger a fresh call to this method, use
        L{rescheduleNextBuild}.

        @param lastActuation: the time of the last actuation, or None for never

        @returns: a Deferred firing with the next time a build should occur (in
        the future), or None for never.
        """
        raise NotImplementedError

    def scheduleNextBuild(self):
        """
        Schedule the next build, re-invoking L{getNextBuildTime}.  This can be
        called at any time, and it will avoid contention with builds being
        started concurrently.

        @returns: Deferred
        """
        return self.actuationLock.run(self._scheduleNextBuild_locked)

    # utilities

    def now(self):
        "Similar to util.now, but patchable by tests"
        return util.now(self._reactor)

    @defer.inlineCallbacks
    def _scheduleNextBuild_locked(self):
        # clear out the existing timer
        if self.actuateAtTimer:
            self.actuateAtTimer.cancel()
        self.actuateAtTimer = None

        # calculate the new time
        actuateAt = yield self.getNextBuildTime(self.lastActuated)

        if actuateAt is None:
            self.actuateAt = None
        else:
            # set up the new timer
            now = self.now()
            self.actuateAt = max(actuateAt, now)
            untilNext = self.actuateAt - now
            if untilNext == 0:
                log.msg(f"{self.__class__.__name__} scheduler <{self.name}>: "
                        "missed scheduled build time - building immediately")
            self.actuateAtTimer = self._reactor.callLater(untilNext,
                                                          self._actuate)

    @defer.inlineCallbacks
    def _actuate(self):
        # called from the timer when it's time to start a build
        self.actuateAtTimer = None
        self.lastActuated = self.actuateAt

        @defer.inlineCallbacks
        def set_state_and_start():
            # bail out if we shouldn't be actuating anymore
            if not self.actuateOk:
                return

            # mark the last build time
            self.actuateAt = None
            yield self.setState('last_build', self.lastActuated)

            try:
                # start the build
                yield self.startBuild()
            except Exception as e:
                log.err(e, 'while actuating')
            finally:
                # schedule the next build (noting the lock is already held)
                yield self._scheduleNextBuild_locked()
        yield self.actuationLock.run(set_state_and_start)


class Periodic(Timed):
    compare_attrs = ('periodicBuildTimer',)

    def __init__(self, name, builderNames, periodicBuildTimer,
                 reason="The Periodic scheduler named '%(name)s' triggered this build",
                 **kwargs):
        super().__init__(name, builderNames, reason=reason, **kwargs)
        if periodicBuildTimer <= 0:
            config.error("periodicBuildTimer must be positive")
        self.periodicBuildTimer = periodicBuildTimer

    def getNextBuildTime(self, lastActuated):
        if lastActuated is None:
            return defer.succeed(self.now())  # meaning "ASAP"
        return defer.succeed(lastActuated + self.periodicBuildTimer)


class NightlyBase(Timed):
    compare_attrs = ('minute', 'hour', 'dayOfMonth', 'month', 'dayOfWeek')

    def __init__(self, name, builderNames, minute=0, hour='*',
                 dayOfMonth='*', month='*', dayOfWeek='*',
                 **kwargs):
        super().__init__(name, builderNames, **kwargs)

        self.minute = minute
        self.hour = hour
        self.dayOfMonth = dayOfMonth
        self.month = month
        self.dayOfWeek = dayOfWeek

    def _timeToCron(self, time, isDayOfWeek=False):
        if isinstance(time, int):
            if isDayOfWeek:
                # Convert from Mon = 0 format to Sun = 0 format for use in
                # croniter
                time = (time + 1) % 7
            return time

        if isinstance(time, str):
            if isDayOfWeek:
                # time could be a comma separated list of values, e.g. "5,sun"
                time_array = str(time).split(',')
                for i, time_val in enumerate(time_array):
                    try:
                        # try to convert value in place
                        # Conversion for croniter (see above)
                        time_array[i] = (int(time_val) + 1) % 7
                    except ValueError:
                        # all non-int values are kept
                        pass
                # Convert the list to a string
                return ','.join([str(s) for s in time_array])

            return time

        if isDayOfWeek:
            # Conversion for croniter (see above)
            time = [(t + 1) % 7 for t in time]

        return ','.join([str(s) for s in time])  # Convert the list to a string

    def getNextBuildTime(self, lastActuated):
        dateTime = lastActuated or self.now()
        sched = (f'{self._timeToCron(self.minute)} {self._timeToCron(self.hour)} '
                 f'{self._timeToCron(self.dayOfMonth)} {self._timeToCron(self.month)} '
                 f'{self._timeToCron(self.dayOfWeek, True)}')
        cron = croniter.croniter(sched, dateTime)
        nextdate = cron.get_next(float)
        return defer.succeed(nextdate)


class Nightly(NightlyBase):

    def __init__(self, name, builderNames, minute=0, hour='*',
                 dayOfMonth='*', month='*', dayOfWeek='*',
                 reason="The Nightly scheduler named '%(name)s' triggered this build",
                 **kwargs):
        super().__init__(name=name, builderNames=builderNames,
                         minute=minute, hour=hour, dayOfMonth=dayOfMonth,
                         month=month, dayOfWeek=dayOfWeek, reason=reason,
                         **kwargs)


@implementer(ITriggerableScheduler)
class NightlyTriggerable(NightlyBase):

    def __init__(self, name, builderNames, minute=0, hour='*',
                 dayOfMonth='*', month='*', dayOfWeek='*',
                 reason="The NightlyTriggerable scheduler named '%(name)s' triggered this build",
                 **kwargs):
        super().__init__(name=name, builderNames=builderNames,
                         minute=minute, hour=hour, dayOfMonth=dayOfMonth,
                         month=month, dayOfWeek=dayOfWeek, reason=reason,
                         **kwargs)

        self._lastTrigger = None

    @defer.inlineCallbacks
    def activate(self):
        yield super().activate()

        if not self.enabled:
            return

        lastTrigger = yield self.getState('lastTrigger', None)
        self._lastTrigger = None
        if lastTrigger:
            try:
                if isinstance(lastTrigger[0], list):
                    self._lastTrigger = (lastTrigger[0],
                                         properties.Properties.fromDict(
                                             lastTrigger[1]),
                                         lastTrigger[2],
                                         lastTrigger[3])
                # handle state from before Buildbot-0.9.0
                elif isinstance(lastTrigger[0], dict):
                    self._lastTrigger = (list(lastTrigger[0].values()),
                                         properties.Properties.fromDict(
                                             lastTrigger[1]),
                                         None,
                                         None)
            except Exception:
                pass
            # If the lastTrigger isn't of the right format, ignore it
            if not self._lastTrigger:
                log.msg(
                    format="NightlyTriggerable Scheduler <%(scheduler)s>: "
                    "could not load previous state; starting fresh",
                    scheduler=self.name)

    def trigger(self, waited_for, sourcestamps=None, set_props=None,
                parent_buildid=None, parent_relationship=None):
        """Trigger this scheduler with the given sourcestamp ID. Returns a
        deferred that will fire when the buildset is finished."""
        assert isinstance(sourcestamps, list), \
            "trigger requires a list of sourcestamps"

        self._lastTrigger = (sourcestamps,
                             set_props,
                             parent_buildid,
                             parent_relationship)

        if set_props:
            propsDict = set_props.asDict()
        else:
            propsDict = {}

        # record the trigger in the db
        d = self.setState('lastTrigger', (sourcestamps,
                                          propsDict,
                                          parent_buildid,
                                          parent_relationship))

        # Trigger expects a callback with the success of the triggered
        # build, if waitForFinish is True.
        # Just return SUCCESS, to indicate that the trigger was successful,
        # don't wait for the nightly to run.
        return (defer.succeed((None, {})), d.addCallback(lambda _: buildstep.SUCCESS))

    @defer.inlineCallbacks
    def startBuild(self):
        if not self.enabled:
            log.msg(format='ignoring build from %(name)s because scheduler '
                           'has been disabled by the user',
                    name=self.name)
            return

        if self._lastTrigger is None:
            return

        (sourcestamps, set_props, parent_buildid,
         parent_relationship) = self._lastTrigger
        self._lastTrigger = None
        yield self.setState('lastTrigger', None)

        # properties for this buildset are composed of our own properties,
        # potentially overridden by anything from the triggering build
        props = properties.Properties()
        props.updateFromProperties(self.properties)
        if set_props:
            props.updateFromProperties(set_props)

        yield self.addBuildsetForSourceStampsWithDefaults(
            reason=self.reason,
            sourcestamps=sourcestamps,
            properties=props,
            parent_buildid=parent_buildid,
            parent_relationship=parent_relationship)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.db import changesources
from buildbot.test.fakedb.base import FakeDBComponent
from buildbot.test.fakedb.row import Row


class ChangeSource(Row):
    table = "changesources"

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]

    def __init__(self, id=None, name='csname', name_hash=None):
        super().__init__(id=id, name=name, name_hash=name_hash)


class ChangeSourceMaster(Row):
    table = "changesource_masters"

    foreignKeys = ('changesourceid', 'masterid')
    required_columns = ('changesourceid', 'masterid')

    def __init__(self, changesourceid=None, masterid=None):
        super().__init__(changesourceid=changesourceid, masterid=masterid)


class FakeChangeSourcesComponent(FakeDBComponent):

    def setUp(self):
        self.changesources = {}
        self.changesource_masters = {}
        self.states = {}

    def insertTestData(self, rows):
        for row in rows:
            if isinstance(row, ChangeSource):
                self.changesources[row.id] = row.name
            if isinstance(row, ChangeSourceMaster):
                self.changesource_masters[row.changesourceid] = row.masterid

    # component methods

    def findChangeSourceId(self, name):
        for cs_id, cs_name in self.changesources.items():
            if cs_name == name:
                return defer.succeed(cs_id)
        new_id = (max(self.changesources) + 1) if self.changesources else 1
        self.changesources[new_id] = name
        return defer.succeed(new_id)

    def getChangeSource(self, changesourceid):
        if changesourceid in self.changesources:
            rv = dict(
                id=changesourceid,
                name=self.changesources[changesourceid],
                masterid=None)
            # only set masterid if the relevant changesource master exists and
            # is active
            rv['masterid'] = self.changesource_masters.get(changesourceid)
            return defer.succeed(rv)
        return None

    def getChangeSources(self, active=None, masterid=None):
        d = defer.DeferredList([
            self.getChangeSource(id) for id in self.changesources
        ])

        @d.addCallback
        def filter(results):
            # filter off the DeferredList results (we know it's good)
            results = [r[1] for r in results]
            # filter for masterid
            if masterid is not None:
                results = [r for r in results
                           if r['masterid'] == masterid]
            # filter for active or inactive if necessary
            if active:
                results = [r for r in results
                           if r['masterid'] is not None]
            elif active is not None:
                results = [r for r in results
                           if r['masterid'] is None]
            return results
        return d

    def setChangeSourceMaster(self, changesourceid, masterid):
        current_masterid = self.changesource_masters.get(changesourceid)
        if current_masterid and masterid is not None and current_masterid != masterid:
            return defer.fail(changesources.ChangeSourceAlreadyClaimedError())
        self.changesource_masters[changesourceid] = masterid
        return defer.succeed(None)

    # fake methods

    def fakeChangeSource(self, name, changesourceid):
        self.changesources[changesourceid] = name

    def fakeChangeSourceMaster(self, changesourceid, masterid):
        if masterid is not None:
            self.changesource_masters[changesourceid] = masterid
        else:
            del self.changesource_masters[changesourceid]

    # assertions

    def assertChangeSourceMaster(self, changesourceid, masterid):
        self.t.assertEqual(self.changesource_masters.get(changesourceid),
                           masterid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from parameterized import parameterized

from mock import Mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.builder import Builder
from buildbot.process.results import CANCELLED
from buildbot.reporters.generators.buildrequest import BuildRequestGenerator
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.reporter import ReporterTestMixin


class TestBuildRequestGenerator(ConfigErrorsMixin, TestReactorMixin,
                                unittest.TestCase, ReporterTestMixin):

    all_messages = ('failing', 'passing', 'warnings', 'exception', 'cancelled')

    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = fakemaster.make_master(self, wantData=True, wantDb=True,
                                             wantMq=True)

        builder = Mock(spec=Builder)
        builder.master = self.master
        self.master.botmaster.getBuilderById = Mock(return_value=builder)

    @parameterized.expand([
        ('tags', 'tag'),
        ('tags', 1),
        ('builders', 'builder'),
        ('builders', 1),
        ('schedulers', 'scheduler'),
        ('schedulers', 1),
        ('branches', 'branch'),
        ('branches', 1),
    ])
    def test_list_params_check_raises(self, arg_name, arg_value):
        kwargs = {arg_name: arg_value}
        g = BuildRequestGenerator(**kwargs)
        with self.assertRaisesConfigError('must be a list or None'):
            g.check()

    def setup_generator(self, message=None, **kwargs):
        if message is None:
            message = {
                "body": "start body",
                "type": "plain",
                "subject": "start subject"
            }

        g = BuildRequestGenerator(**kwargs)

        g.formatter = Mock(spec=g.formatter)
        g.formatter.format_message_for_build.return_value = message

        return g

    @defer.inlineCallbacks
    def test_build_message_start_no_result(self):
        g = yield self.setup_generator()
        buildrequest = yield self.insert_buildrequest_new()
        build = yield g.partial_build_dict(self.master, buildrequest)
        report = yield g.buildrequest_message(self.master, build)

        g.formatter.format_message_for_build.assert_called_with(self.master, build,
                                                                is_buildset=True,
                                                                mode=self.all_messages,
                                                                users=[])

        self.assertEqual(report, {
            'body': 'start body',
            'subject': 'start subject',
            'type': 'plain',
            'results': None,
            'builds': [build],
            'users': [],
            'patches': [],
            'logs': []
        })

    @defer.inlineCallbacks
    def test_build_message_add_patch(self):
        g = yield self.setup_generator(add_patch=True)
        buildrequest = yield self.insert_buildrequest_new(insert_patch=True)
        build = yield g.partial_build_dict(self.master, buildrequest)
        report = yield g.buildrequest_message(self.master, build)

        patch_dict = {
            'author': 'him@foo',
            'body': b'hello, world',
            'comment': 'foo',
            'level': 3,
            'patchid': 99,
            'subdir': '/foo'
        }
        self.assertEqual(report['patches'], [patch_dict])

    @defer.inlineCallbacks
    def test_build_message_add_patch_no_patch(self):
        g = yield self.setup_generator(add_patch=True)
        buildrequest = yield self.insert_buildrequest_new(insert_patch=False)
        build = yield g.partial_build_dict(self.master, buildrequest)
        report = yield g.buildrequest_message(self.master, build)
        self.assertEqual(report['patches'], [])

    @defer.inlineCallbacks
    def test_generate_new(self):
        g = yield self.setup_generator(add_patch=True)
        buildrequest = yield self.insert_buildrequest_new(insert_patch=False)
        build = yield g.partial_build_dict(self.master, buildrequest)
        report = yield g.generate(self.master, None, ('buildrequests', 11, 'new'), buildrequest)

        self.assertEqual(report, {
            'body': 'start body',
            'subject': 'start subject',
            'type': 'plain',
            'results': None,
            'builds': [build],
            'users': [],
            'patches': [],
            'logs': []
        })

    @defer.inlineCallbacks
    def test_generate_cancel(self):
        self.maxDiff = None
        g = yield self.setup_generator(add_patch=True)
        buildrequest = yield self.insert_buildrequest_new(insert_patch=False)
        build = yield g.partial_build_dict(self.master, buildrequest)
        report = yield g.generate(self.master, None, ('buildrequests', 11, 'cancel'), buildrequest)

        build['complete'] = True
        build['results'] = CANCELLED

        self.assertEqual(report, {
            'body': 'start body',
            'subject': 'start subject',
            'type': 'plain',
            'results': CANCELLED,
            'builds': [build],
            'users': [],
            'patches': [],
            'logs': []
        })

    @defer.inlineCallbacks
    def test_generate_none(self):
        g = BuildRequestGenerator(builders=['not_existing_builder'])
        buildrequest = yield self.insert_buildrequest_new()
        report = yield g.generate(self.master, None, ('buildrequests', 11, 'new'), buildrequest)

        self.assertIsNone(report, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from sqlalchemy.engine import url
from sqlalchemy.pool import NullPool

from twisted.python import runtime
from twisted.trial import unittest

from buildbot.db import enginestrategy


class BuildbotCreateEngineTest(unittest.TestCase):

    "Test the special case methods, without actually creating a db"

    # used several times below
    mysql_kwargs = dict(
        basedir='my-base-dir',
        connect_args=dict(init_command='SET default_storage_engine=MyISAM'),
        pool_recycle=3600)
    sqlite_kwargs = dict(basedir='/my-base-dir', poolclass=NullPool)

    # utility

    def filter_kwargs(self, kwargs):
        # filter out the listeners list to just include the class name
        if 'listeners' in kwargs:
            kwargs['listeners'] = [lstnr.__class__.__name__
                                   for lstnr in kwargs['listeners']]
        return kwargs

    # tests

    def test_sqlite_pct_sub(self):
        u = url.make_url("sqlite:///%(basedir)s/x/state.sqlite")
        kwargs = dict(basedir='/my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["sqlite:////my-base-dir/x/state.sqlite", 1,
                          self.sqlite_kwargs])

    def test_sqlite_relpath(self):
        url_src = "sqlite:///x/state.sqlite"
        basedir = "/my-base-dir"
        expected_url = "sqlite:////my-base-dir/x/state.sqlite"

        # this looks a whole lot different on windows
        if runtime.platformType == 'win32':
            url_src = r'sqlite:///X\STATE.SQLITE'
            basedir = r'C:\MYBASE~1'
            expected_url = r'sqlite:///C:\MYBASE~1\X\STATE.SQLITE'

        exp_kwargs = self.sqlite_kwargs.copy()
        exp_kwargs['basedir'] = basedir

        u = url.make_url(url_src)
        kwargs = dict(basedir=basedir)
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         [expected_url, 1, exp_kwargs])

    def test_sqlite_abspath(self):
        u = url.make_url("sqlite:////x/state.sqlite")
        kwargs = dict(basedir='/my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["sqlite:////x/state.sqlite", 1, self.sqlite_kwargs])

    def test_sqlite_memory(self):
        u = url.make_url("sqlite://")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_sqlite(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["sqlite://", 1,  # only one conn at a time
                          dict(basedir='my-base-dir',
                               connect_args=dict(check_same_thread=False))])

    def test_mysql_simple(self):
        u = url.make_url("mysql://host/dbname")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql://host/dbname?charset=utf8&use_unicode=True", None,
                          self.mysql_kwargs])

    def test_mysql_userport(self):
        u = url.make_url("mysql://user:pass@host:1234/dbname")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql://user:pass@host:1234/dbname?"
                          "charset=utf8&use_unicode=True", None, self.mysql_kwargs])

    def test_mysql_local(self):
        u = url.make_url("mysql:///dbname")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&use_unicode=True", None,
                          self.mysql_kwargs])

    def test_mysql_args(self):
        u = url.make_url("mysql:///dbname?foo=bar")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&foo=bar&use_unicode=True",
                          None, self.mysql_kwargs])

    def test_mysql_max_idle(self):
        u = url.make_url("mysql:///dbname?max_idle=1234")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        exp = self.mysql_kwargs.copy()
        exp['pool_recycle'] = 1234
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&use_unicode=True", None,
                          exp])

    def test_mysql_good_charset(self):
        u = url.make_url("mysql:///dbname?charset=utf8")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&use_unicode=True", None,
                          self.mysql_kwargs])

    def test_mysql_bad_charset(self):
        u = url.make_url("mysql:///dbname?charset=ebcdic")
        kwargs = dict(basedir='my-base-dir')
        with self.assertRaises(TypeError):
            enginestrategy.special_case_mysql(u, kwargs)

    def test_mysql_good_use_unicode(self):
        u = url.make_url("mysql:///dbname?use_unicode=True")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&use_unicode=True", None,
                          self.mysql_kwargs])

    def test_mysql_bad_use_unicode(self):
        u = url.make_url("mysql:///dbname?use_unicode=maybe")
        kwargs = dict(basedir='my-base-dir')
        with self.assertRaises(TypeError):
            enginestrategy.special_case_mysql(u, kwargs)

    def test_mysql_storage_engine(self):
        u = url.make_url("mysql:///dbname?storage_engine=foo")
        kwargs = dict(basedir='my-base-dir')
        u, kwargs, max_conns = enginestrategy.special_case_mysql(u, kwargs)
        exp = self.mysql_kwargs.copy()
        exp['connect_args'] = dict(
            init_command='SET default_storage_engine=foo')
        self.assertEqual([str(u), max_conns, self.filter_kwargs(kwargs)],
                         ["mysql:///dbname?charset=utf8&use_unicode=True", None,
                          exp])


class BuildbotEngineStrategy(unittest.TestCase):

    "Test create_engine by creating a sqlite in-memory db"

    def test_create_engine(self):
        engine = enginestrategy.create_engine('sqlite://', basedir="/base")
        self.assertEqual(engine.scalar("SELECT 13 + 14"), 27)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import test_results
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces


class TestResultsEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = test_results.TestResultsEndpoint
    resourceTypeClass = test_results.TestResult

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88,
                         builderid=88, workerid=47),
            fakedb.Step(id=131, number=132, name='step132', buildid=30),
            fakedb.TestResultSet(id=13, builderid=88, buildid=30, stepid=131, description='desc',
                                 category='cat', value_unit='ms', complete=1),
            fakedb.TestName(id=301, builderid=88, name='name301'),
            fakedb.TestName(id=302, builderid=88, name='name302'),
            fakedb.TestCodePath(id=401, builderid=88, path='path401'),
            fakedb.TestCodePath(id=402, builderid=88, path='path402'),
            fakedb.TestResult(id=101, builderid=88, test_result_setid=13, line=400, value='v101'),
            fakedb.TestResult(id=102, builderid=88, test_result_setid=13,
                              test_nameid=301, test_code_pathid=401, line=401, value='v102'),
            fakedb.TestResult(id=103, builderid=88, test_result_setid=13,
                              test_nameid=302, test_code_pathid=402, line=402,
                              duration_ns=1012, value='v103'),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_existing_results(self):
        results = yield self.callGet(('test_result_sets', 13, 'results'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_resultid'] for r in results], [101, 102, 103])

    @defer.inlineCallbacks
    def test_get_missing_results(self):
        results = yield self.callGet(('test_result_sets', 14, 'results'))
        self.assertEqual(results, [])


class TestResult(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = test_results.TestResult(self.master)

    def test_signature_add_test_results(self):
        @self.assertArgSpecMatches(self.master.data.updates.addTestResults,
                                   self.rtype.addTestResults)
        def addTestResults(self, builderid, test_result_setid, result_values):
            pass

    @defer.inlineCallbacks
    def test_add_test_results(self):
        result_values = [
            {'test_name': 'name1', 'value': '1'},
            {'test_name': 'name2', 'duration_ns': 1000, 'value': '1'},
            {'test_name': 'name3', 'test_code_path': 'path2', 'value': '2'},
            {'test_name': 'name4', 'test_code_path': 'path3', 'value': '3'},
            {'test_name': 'name5', 'test_code_path': 'path4', 'line': 4, 'value': '4'},
            {'test_code_path': 'path5', 'line': 5, 'value': '5'},
        ]
        yield self.rtype.addTestResults(builderid=88, test_result_setid=13,
                                        result_values=result_values)

        self.master.mq.assertProductions([])

        results = yield self.master.db.test_results.getTestResults(builderid=88,
                                                                   test_result_setid=13)
        resultid = results[0]['id']
        self.assertEqual(results, [
            {'id': resultid, 'builderid': 88, 'test_result_setid': 13, 'test_name': 'name1',
             'test_code_path': None, 'line': None, 'duration_ns': None, 'value': '1'},
            {'id': resultid + 1, 'builderid': 88, 'test_result_setid': 13, 'test_name': 'name2',
             'test_code_path': None, 'line': None, 'duration_ns': 1000, 'value': '1'},
            {'id': resultid + 2, 'builderid': 88, 'test_result_setid': 13, 'test_name': 'name3',
             'test_code_path': 'path2', 'line': None, 'duration_ns': None, 'value': '2'},
            {'id': resultid + 3, 'builderid': 88, 'test_result_setid': 13, 'test_name': 'name4',
             'test_code_path': 'path3', 'line': None, 'duration_ns': None, 'value': '3'},
            {'id': resultid + 4, 'builderid': 88, 'test_result_setid': 13, 'test_name': 'name5',
             'test_code_path': 'path4', 'line': 4, 'duration_ns': None, 'value': '4'},
            {'id': resultid + 5, 'builderid': 88, 'test_result_setid': 13, 'test_name': None,
             'test_code_path': 'path5', 'line': 5, 'duration_ns': None, 'value': '5'},
        ])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log

from buildbot.util.eventual import fireEventually
from buildbot.warnings import warn_deprecated
from buildbot.worker.protocols import base


class Listener(base.Listener):
    pass


class ProxyMixin():

    def __init__(self, impl):
        assert isinstance(impl, self.ImplClass)
        self.impl = impl
        self._disconnect_listeners = []

    def callRemote(self, message, *args, **kw):
        method = getattr(self.impl, f"remote_{message}", None)
        if method is None:
            raise AttributeError(f"No such method: remote_{message}")
        try:
            state = method(*args, **kw)
        except TypeError:
            log.msg(f"{method} didn't accept {args} and {kw}")
            raise
        # break callback recursion for large transfers by using fireEventually
        return fireEventually(state)

    def notifyOnDisconnect(self, cb):
        pass

    def dontNotifyOnDisconnect(self, cb):
        pass


# just add ProxyMixin capability to the RemoteCommandProxy
# so that callers of callRemote actually directly call the proper method
class RemoteCommandProxy(ProxyMixin):
    ImplClass = base.RemoteCommandImpl


class FileReaderProxy(ProxyMixin):
    ImplClass = base.FileReaderImpl


class FileWriterProxy(ProxyMixin):
    ImplClass = base.FileWriterImpl


class Connection(base.Connection):
    proxies = {base.FileWriterImpl: FileWriterProxy,
               base.FileReaderImpl: FileReaderProxy}

    def __init__(self, master_or_worker, worker=None):
        # All the existing code passes just the name to the Connection, however we'll need to
        # support an older versions of buildbot-worker using two parameter signature for some time.
        if worker is None:
            worker = master_or_worker
        else:
            warn_deprecated('3.2.0', 'LocalWorker: Using different version of buildbot-worker ' +
                            'than buildbot is not supported')

        super().__init__(worker.workername)
        self.worker = worker

    def loseConnection(self):
        self.notifyDisconnected()

    def remotePrint(self, message):
        return defer.maybeDeferred(self.worker.bot.remote_print, message)

    def remoteGetWorkerInfo(self):
        return defer.maybeDeferred(self.worker.bot.remote_getWorkerInfo)

    def remoteSetBuilderList(self, builders):
        return defer.maybeDeferred(self.worker.bot.remote_setBuilderList, builders)

    def remoteStartCommand(self, remoteCommand, builderName, commandId, commandName, args):
        remoteCommand = RemoteCommandProxy(remoteCommand)
        args = self.createArgsProxies(args)
        workerforbuilder = self.worker.bot.builders[builderName]
        return defer.maybeDeferred(workerforbuilder.remote_startCommand, remoteCommand,
                                   commandId, commandName, args)

    def remoteShutdown(self):
        return defer.maybeDeferred(self.worker.stopService)

    def remoteStartBuild(self, builderName):
        return defer.succeed(self.worker.bot.builders[builderName].remote_startBuild())

    def remoteInterruptCommand(self, builderName, commandId, why):
        workerforbuilder = self.worker.bot.builders[builderName]
        return defer.maybeDeferred(workerforbuilder.remote_interruptCommand, commandId, why)

    def get_peer(self):
        return "local"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re


class RevlinkMatch:

    def __init__(self, repo_urls, revlink):
        if isinstance(repo_urls, str):
            repo_urls = [repo_urls]
        self.repo_urls = [re.compile(url) for url in repo_urls]
        self.revlink = revlink

    def __call__(self, rev, repo):
        for url in self.repo_urls:
            m = url.match(repo)
            if m:
                return m.expand(self.revlink) % rev
        return None


GithubRevlink = RevlinkMatch(
    repo_urls=[r'https://github.com/([^/]*)/([^/]*?)(?:\.git)?$',
               r'git://github.com/([^/]*)/([^/]*?)(?:\.git)?$',
               r'git@github.com:([^/]*)/([^/]*?)(?:\.git)?$',
               r'ssh://git@github.com/([^/]*)/([^/]*?)(?:\.git)?$'
               ],
    revlink=r'https://github.com/\1/\2/commit/%s')


BitbucketRevlink = RevlinkMatch(
        repo_urls=[r'https://[^@]*@bitbucket.org/([^/]*)/([^/]*?)(?:\.git)?$',
                   r'git@bitbucket.org:([^/]*)/([^/]*?)(?:\.git)?$'],
        revlink=r'https://bitbucket.org/\1/\2/commits/%s')


class GitwebMatch(RevlinkMatch):

    def __init__(self, repo_urls, revlink):
        super().__init__(repo_urls=repo_urls,
                         revlink=revlink + r'?p=\g<repo>;a=commit;h=%s')


SourceforgeGitRevlink = GitwebMatch(
    repo_urls=[r'^git://([^.]*).git.sourceforge.net/gitroot/(?P<repo>.*)$',
               r'[^@]*@([^.]*).git.sourceforge.net:gitroot/(?P<repo>.*)$',
               r'ssh://(?:[^@]*@)?([^.]*).git.sourceforge.net/gitroot/(?P<repo>.*)$',
               ],
    revlink=r'http://\1.git.sourceforge.net/git/gitweb.cgi')

# SourceForge recently upgraded to another platform called Allura
# See introduction:
# https://sourceforge.net/p/forge/documentation/Classic%20vs%20New%20SourceForge%20projects/
# And as reference:
# https://sourceforge.net/p/forge/community-docs/SVN%20and%20project%20upgrades/
SourceforgeGitRevlink_AlluraPlatform = RevlinkMatch(
    repo_urls=[r'git://git.code.sf.net/p/(?P<repo>.*)$',
               r'http://git.code.sf.net/p/(?P<repo>.*)$',
               r'ssh://(?:[^@]*@)?git.code.sf.net/p/(?P<repo>.*)$'
               ],
    revlink=r'https://sourceforge.net/p/\1/ci/%s/')


class RevlinkMultiplexer:

    def __init__(self, *revlinks):
        self.revlinks = revlinks

    def __call__(self, rev, repo):
        for revlink in self.revlinks:
            url = revlink(rev, repo)
            if url:
                return url
        return None


default_revlink_matcher = RevlinkMultiplexer(GithubRevlink,
                                             BitbucketRevlink,
                                             SourceforgeGitRevlink,
                                             SourceforgeGitRevlink_AlluraPlatform)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import stat

from twisted.internet import defer
from twisted.python import log
from twisted.python.reflect import namedModule

from buildbot.pbutil import decode
from buildbot.process import remotecommand
from buildbot.util import deferwaiter
from buildbot.util import path_expand_user
from buildbot.worker.protocols import base


class Listener(base.UpdateRegistrationListener):
    name = "MsgPackListener"

    def __init__(self, master):
        super().__init__()
        self.ConnectionClass = Connection
        self.master = master

    def get_manager(self):
        return self.master.msgmanager

    def before_connection_setup(self, protocol, workerName):
        log.msg(f"worker '{workerName}' attaching")


class BasicRemoteCommand():
    # only has basic functions needed for remoteSetBuilderList in class Connection
    # when waiting for update messages
    def __init__(self, worker_name, expected_keys, error_msg):
        self.worker_name = worker_name
        self.update_results = {}
        self.expected_keys = expected_keys
        self.error_msg = error_msg
        self.d = defer.Deferred()

    def wait_until_complete(self):
        return self.d

    def remote_update_msgpack(self, args):
        # args is a list of tuples
        # first element of the tuple is a key, second element is a value
        for key, value in args:
            if key not in self.update_results:
                self.update_results[key] = value

    def remote_complete(self, args):
        if 'rc' not in self.update_results:
            self.d.errback(Exception(f"Worker {self.worker_name} reconfiguration or connection to "
                                     f"master failed. {self.error_msg}. 'rc' did not arrive."))
            return

        if self.update_results['rc'] != 0:
            self.d.errback(Exception(f"Worker {self.worker_name} reconfiguration or connection to "
                                     f"master failed. {self.error_msg}. Error number: "
                                     f"{self.update_results['rc']}"))
            return

        for key in self.expected_keys:
            if key not in self.update_results:
                self.d.errback(Exception(f"Worker {self.worker_name} reconfiguration or connection "
                                         f"to master failed. {self.error_msg} "
                                         f"Key '{key}' is missing."))
                return

        self.d.callback(None)


class Connection(base.Connection):
    # TODO: configure keepalive_interval in
    # c['protocols']['msgpack']['keepalive_interval']
    keepalive_timer = None
    keepalive_interval = 3600
    info = None

    def __init__(self, master, worker, protocol):
        super().__init__(worker.workername)
        self.master = master
        self.worker = worker
        self.protocol = protocol
        self._keepalive_waiter = deferwaiter.DeferWaiter()
        self._keepalive_action_handler = \
            deferwaiter.RepeatedActionHandler(master.reactor, self._keepalive_waiter,
                                              self.keepalive_interval, self._do_keepalive)

    # methods called by the BuildbotWebSocketServerProtocol

    @defer.inlineCallbacks
    def attached(self, protocol):
        self.startKeepaliveTimer()
        self.notifyOnDisconnect(self._stop_keepalive_timer)
        yield self.worker.attached(self)

    def detached(self, protocol):
        self.stopKeepaliveTimer()
        self.protocol = None
        self.notifyDisconnected()

    # disconnection handling
    @defer.inlineCallbacks
    def _stop_keepalive_timer(self):
        self.stopKeepaliveTimer()
        yield self._keepalive_waiter.wait()

    def loseConnection(self):
        self.stopKeepaliveTimer()
        self.protocol.transport.abortConnection()

    # keepalive handling

    def _do_keepalive(self):
        return self.remoteKeepalive()

    def stopKeepaliveTimer(self):
        self._keepalive_action_handler.stop()

    def startKeepaliveTimer(self):
        assert self.keepalive_interval
        self._keepalive_action_handler.start()

    # methods to send messages to the worker

    def remoteKeepalive(self):
        return self.protocol.get_message_result({'op': 'keepalive'})

    def remotePrint(self, message):
        return self.protocol.get_message_result({'op': 'print', 'message': message})

    @defer.inlineCallbacks
    def remoteGetWorkerInfo(self):
        info = yield self.protocol.get_message_result({'op': 'get_worker_info'})
        self.info = decode(info)

        worker_system = self.info.get("system", None)
        if worker_system == "nt":
            self.path_module = namedModule("ntpath")
            self.path_expanduser = path_expand_user.nt_expanduser
        else:
            # most everything accepts / as separator, so posix should be a reasonable fallback
            self.path_module = namedModule("posixpath")
            self.path_expanduser = path_expand_user.posix_expanduser
        return self.info

    def _set_worker_settings(self):
        # the lookahead here (`(?=.)`) ensures that `\r` doesn't match at the end
        # of the buffer
        # we also convert cursor control sequence to newlines
        # and ugly \b+ (use of backspace to implement progress bar)
        newline_re = r'(\r\n|\r(?=.)|\033\[u|\033\[[0-9]+;[0-9]+[Hf]|\033\[2J|\x08+)'
        return self.protocol.get_message_result({
            'op': 'set_worker_settings',
            'args': {'newline_re': newline_re,
                     'max_line_length': 4096,
                     'buffer_timeout': 5,
                     'buffer_size': 64 * 1024}})

    def create_remote_command(self, worker_name, expected_keys, error_msg):
        command_id = remotecommand.RemoteCommand.generate_new_command_id()
        command = BasicRemoteCommand(worker_name, expected_keys, error_msg)
        self.protocol.command_id_to_command_map[command_id] = command
        return (command, command_id)

    @defer.inlineCallbacks
    def remoteSetBuilderList(self, builders):
        yield self._set_worker_settings()

        basedir = self.info['basedir']
        builder_names = [name for name, _ in builders]
        self.builder_basedirs = {name: self.path_module.join(basedir, builddir)
                                 for name, builddir in builders}

        wanted_dirs = {builddir for _, builddir in builders}
        wanted_dirs.add('info')
        dirs_to_mkdir = set(wanted_dirs)
        command, command_id = \
            self.create_remote_command(self.worker.workername, ['files'],
                                       'Worker could not send a list of builder directories.')

        yield self.protocol.get_message_result({'op': 'start_command',
                                                'command_id': command_id,
                                                'command_name': 'listdir',
                                                'args': {'path': basedir}})

        # wait until command is over to get the update request message with args['files']
        yield command.wait_until_complete()
        files = command.update_results['files']

        paths_to_rmdir = []

        for dir in files:
            dirs_to_mkdir.discard(dir)
            if dir not in wanted_dirs:
                if self.info['delete_leftover_dirs']:
                    # send 'stat' start_command and wait for status information which comes from
                    # worker in a response message. Status information is saved in update_results
                    # dictionary with key 'stat'. 'stat' value is a tuple of 10 elements, where
                    # first element is File mode. It goes to S_ISDIR(mode) to check if path is
                    # a directory so that files are not deleted
                    path = self.path_module.join(basedir, dir)
                    command, command_id = \
                        self.create_remote_command(self.worker.workername, ['stat'],
                                                   "Worker could not send status " +
                                                   "information about its files.")
                    yield self.protocol.get_message_result({'op': 'start_command',
                                                            'command_id': command_id,
                                                            'command_name': 'stat',
                                                            'args': {'path': path}})
                    yield command.wait_until_complete()
                    mode = command.update_results['stat'][0]
                    if stat.S_ISDIR(mode):
                        paths_to_rmdir.append(path)

        if paths_to_rmdir:
            log.msg(f"Deleting directory '{paths_to_rmdir}' that is not being "
                    "used by the buildmaster.")

            # remove leftover directories from worker
            command, command_id = \
                self.create_remote_command(self.worker.workername, [],
                                           "Worker could not remove directories.")
            yield self.protocol.get_message_result({'op': 'start_command',
                                                    'command_id': command_id,
                                                    'command_name': 'rmdir',
                                                    'args': {'paths': paths_to_rmdir}})
            yield command.wait_until_complete()

        paths_to_mkdir = [self.path_module.join(basedir, dir)
                          for dir in sorted(list(dirs_to_mkdir))]
        if paths_to_mkdir:
            # make wanted builder directories which do not exist in worker yet
            command, command_id = self.create_remote_command(self.worker.workername, [],
                                                             "Worker could not make directories.")
            yield self.protocol.get_message_result({'op': 'start_command',
                                                    'command_id': command_id,
                                                    'command_name': 'mkdir',
                                                    'args': {'paths': paths_to_mkdir}})
            yield command.wait_until_complete()

        self.builders = builder_names
        return builder_names

    @defer.inlineCallbacks
    def remoteStartCommand(self, remoteCommand, builderName, commandId, commandName, args):
        if commandName == "mkdir":
            if isinstance(args['dir'], list):
                args['paths'] = [self.path_module.join(self.builder_basedirs[builderName], dir)
                                 for dir in args['dir']]
            else:
                args['paths'] = [self.path_module.join(self.builder_basedirs[builderName],
                                                       args['dir'])]
            del args['dir']

        if commandName == "rmdir":
            if isinstance(args['dir'], list):
                args['paths'] = [self.path_module.join(self.builder_basedirs[builderName], dir)
                                 for dir in args['dir']]
            else:
                args['paths'] = [self.path_module.join(self.builder_basedirs[builderName],
                                                       args['dir'])]
            del args['dir']

        if commandName == "cpdir":
            args['from_path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                      args['fromdir'])
            args['to_path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                    args['todir'])
            del args['fromdir']
            del args['todir']

        if commandName == "stat":
            args['path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                 args.get('workdir', ''), args['file'])
            del args['file']

        if commandName == "glob":
            args['path'] = self.path_module.join(self.builder_basedirs[builderName], args['path'])

        if commandName == "listdir":
            args['path'] = self.path_module.join(self.builder_basedirs[builderName], args['dir'])
            del args['dir']

        if commandName == "rmfile":
            args['path'] = self.path_module.join(self.builder_basedirs[builderName], args['path'])

        if commandName == "shell":
            args['workdir'] = self.path_module.join(self.builder_basedirs[builderName],
                                                    args['workdir'])

        if commandName == "uploadFile":
            commandName = "upload_file"
            args['path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                 args['workdir'],
                                                 self.path_expanduser(args['workersrc'],
                                                                      self.info['environ']))

        if commandName == "uploadDirectory":
            commandName = "upload_directory"
            args['path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                 args['workdir'],
                                                 self.path_expanduser(args['workersrc'],
                                                                      self.info['environ']))

        if commandName == "downloadFile":
            commandName = "download_file"
            args['path'] = self.path_module.join(self.builder_basedirs[builderName],
                                                 args['workdir'],
                                                 self.path_expanduser(args['workerdest'],
                                                                      self.info['environ']))
        if "want_stdout" in args:
            if args["want_stdout"] == 1:
                args["want_stdout"] = True
            else:
                args["want_stdout"] = False

        if "want_stderr" in args:
            if args["want_stderr"] == 1:
                args["want_stderr"] = True
            else:
                args["want_stderr"] = False

        self.protocol.command_id_to_command_map[commandId] = remoteCommand
        if 'reader' in args:
            self.protocol.command_id_to_reader_map[commandId] = args['reader']
            del args['reader']
        if 'writer' in args:
            self.protocol.command_id_to_writer_map[commandId] = args['writer']
            del args['writer']
        yield self.protocol.get_message_result({'op': 'start_command', 'builder_name': builderName,
                                               'command_id': commandId, 'command_name': commandName,
                                               'args': args})

    @defer.inlineCallbacks
    def remoteShutdown(self):
        yield self.protocol.get_message_result({'op': 'shutdown'})

    def remoteStartBuild(self, builderName):
        pass

    @defer.inlineCallbacks
    def remoteInterruptCommand(self, builderName, commandId, why):
        yield self.protocol.get_message_result({'op': 'interrupt_command',
                                               'builder_name': builderName, 'command_id': commandId,
                                               'why': why})

    # perspective methods called by the worker

    def perspective_keepalive(self):
        self.worker.messageReceivedFromWorker()

    def perspective_shutdown(self):
        self.worker.messageReceivedFromWorker()
        self.worker.shutdownRequested()

    def get_peer(self):
        p = self.protocol.transport.getPeer()
        return f"{p.host}:{p.port}"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import sys
import traceback

from twisted.internet import defer

from buildbot.clients import sendchange as sendchange_client
from buildbot.util import in_reactor


@in_reactor
@defer.inlineCallbacks
def sendchange(config):
    encoding = config.get('encoding', 'utf8')
    who = config.get('who')
    auth = config.get('auth')
    master = config.get('master')
    branch = config.get('branch')
    category = config.get('category')
    revision = config.get('revision')
    properties = config.get('properties', {})
    repository = config.get('repository', '')
    vc = config.get('vc', None)
    project = config.get('project', '')
    revlink = config.get('revlink', '')
    when = config.get('when')
    comments = config.get('comments')
    files = config.get('files', ())
    codebase = config.get('codebase', None)

    s = sendchange_client.Sender(master, auth, encoding=encoding)
    try:
        yield s.send(branch, revision, comments, files, who=who,
                     category=category, when=when, properties=properties,
                     repository=repository, vc=vc, project=project, revlink=revlink,
                     codebase=codebase)
    except Exception:
        print("change not sent:")
        traceback.print_exc(file=sys.stdout)
        return 1
    else:
        print("change sent successfully")
        return 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json
import os
import shutil
import tempfile

from mock import Mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.process import remotetransfer
from buildbot.process.properties import Interpolate
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import SKIPPED
from buildbot.process.results import SUCCESS
from buildbot.steps import transfer
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectGlob
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectStat
from buildbot.test.steps import ExpectUploadDirectory
from buildbot.test.steps import ExpectUploadFile
from buildbot.test.steps import TestBuildStepMixin


class TestFileUpload(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        fd, self.destfile = tempfile.mkstemp()
        os.close(fd)
        os.unlink(self.destfile)
        return self.setup_test_build_step()

    def tearDown(self):
        if os.path.exists(self.destfile):
            os.unlink(self.destfile)
        return self.tear_down_test_build_step()

    def testConstructorModeType(self):
        with self.assertRaises(config.ConfigErrors):
            transfer.FileUpload(workersrc=__file__,
                                masterdest='xyz', mode='g+rwx')

    def testBasic(self):
        self.setup_step(
            transfer.FileUpload(workersrc='srcfile', masterdest=self.destfile))

        self.expect_commands(
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="uploading srcfile")
        d = self.run_step()
        return d

    def testWorker2_16(self):
        self.setup_step(
            transfer.FileUpload(workersrc='srcfile', masterdest=self.destfile),
            worker_version={'*': '2.16'})

        self.expect_commands(
            ExpectUploadFile(slavesrc="srcfile", workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="uploading srcfile")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def testTimestamp(self):
        self.setup_step(
            transfer.FileUpload(workersrc=__file__, masterdest=self.destfile, keepstamp=True))

        timestamp = (os.path.getatime(__file__),
                     os.path.getmtime(__file__))

        self.expect_commands(
            ExpectUploadFile(workersrc=__file__, workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=True,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string('test\n', timestamp=timestamp)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string=f"uploading {os.path.basename(__file__)}"
            )

        yield self.run_step()

        desttimestamp = (os.path.getatime(self.destfile),
                         os.path.getmtime(self.destfile))

        srctimestamp = [int(t) for t in timestamp]
        desttimestamp = [int(d) for d in desttimestamp]

        self.assertEqual(srctimestamp[0], desttimestamp[0])
        self.assertEqual(srctimestamp[1], desttimestamp[1])

    def testDescriptionDone(self):
        self.setup_step(
            transfer.FileUpload(workersrc=__file__, masterdest=self.destfile,
                                url="http://server/file", descriptionDone="Test File Uploaded"))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectUploadFile(workersrc=__file__, workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string="Test File Uploaded")

        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def testURL(self):
        self.setup_step(transfer.FileUpload(workersrc=__file__, masterdest=self.destfile,
                                           url="http://server/file"))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectUploadFile(workersrc=__file__, workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string=f"uploading {os.path.basename(__file__)}"
            )

        yield self.run_step()

        self.step.addURL.assert_called_once_with(
            os.path.basename(self.destfile), "http://server/file")

    @defer.inlineCallbacks
    def testURLText(self):
        self.setup_step(transfer.FileUpload(workersrc=__file__,
                                           masterdest=self.destfile, url="http://server/file",
                                           urlText="testfile"))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectUploadFile(workersrc=__file__, workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string=f"uploading {os.path.basename(__file__)}"
            )

        yield self.run_step()

        self.step.addURL.assert_called_once_with(
            "testfile", "http://server/file")

    def testFailure(self):
        self.setup_step(
            transfer.FileUpload(workersrc='srcfile', masterdest=self.destfile))

        self.expect_commands(
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .exit(1))

        self.expect_outcome(
            result=FAILURE,
            state_string="uploading srcfile (failure)")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def testException(self):
        self.setup_step(
            transfer.FileUpload(workersrc='srcfile', masterdest=self.destfile))

        writers = []

        self.expect_commands(
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=262144, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n", out_writers=writers, error=RuntimeError('uh oh')))

        self.expect_outcome(
            result=EXCEPTION, state_string="uploading srcfile (exception)")
        yield self.run_step()

        self.assertEqual(len(writers), 1)
        self.assertEqual(writers[0].cancel.called, True)

        self.assertEqual(
            len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def test_interrupt(self):
        self.setup_step(transfer.FileUpload(workersrc='srcfile', masterdest=self.destfile))

        self.expect_commands(
            ExpectUploadFile(workersrc='srcfile', workdir='wkdir', blocksize=262144,
                             maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter),
                             interrupted=True)
            .exit(0))

        self.interrupt_nth_remote_command(0)

        self.expect_outcome(result=CANCELLED,
                           state_string="uploading srcfile (cancelled)")
        self.expect_log_file('interrupt', 'interrupt reason')
        yield self.run_step()

    def test_init_workersrc_keyword(self):
        step = transfer.FileUpload(
            workersrc='srcfile', masterdest='dstfile')

        self.assertEqual(step.workersrc, 'srcfile')

    def test_init_workersrc_positional(self):
        step = transfer.FileUpload('srcfile', 'dstfile')

        self.assertEqual(step.workersrc, 'srcfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.FileUpload()
        with self.assertRaises(TypeError):
            transfer.FileUpload('src')


class TestDirectoryUpload(TestBuildStepMixin, TestReactorMixin,
                          unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.destdir = os.path.abspath('destdir')
        if os.path.exists(self.destdir):
            shutil.rmtree(self.destdir)

        return self.setup_test_build_step()

    def tearDown(self):
        if os.path.exists(self.destdir):
            shutil.rmtree(self.destdir)

        return self.tear_down_test_build_step()

    def testBasic(self):
        self.setup_step(
            transfer.DirectoryUpload(workersrc="srcdir", masterdest=self.destdir))

        self.expect_commands(
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading srcdir")
        d = self.run_step()
        return d

    def testWorker2_16(self):
        self.setup_step(
            transfer.DirectoryUpload(
                workersrc="srcdir", masterdest=self.destdir),
            worker_version={'*': '2.16'})

        self.expect_commands(
            ExpectUploadDirectory(slavesrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading srcdir")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def test_url(self):
        self.setup_step(transfer.DirectoryUpload(workersrc="srcdir", masterdest=self.destdir,
                                                url="http://server/dir"))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectUploadDirectory(workersrc='srcdir', workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading srcdir")

        yield self.run_step()

        self.step.addURL.assert_called_once_with("destdir", "http://server/dir")

    @defer.inlineCallbacks
    def test_url_text(self):
        self.setup_step(transfer.DirectoryUpload(workersrc="srcdir", masterdest=self.destdir,
                                                url="http://server/dir", urlText='url text'))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectUploadDirectory(workersrc='srcdir', workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading srcdir")

        yield self.run_step()

        self.step.addURL.assert_called_once_with("url text", "http://server/dir")

    @defer.inlineCallbacks
    def testFailure(self):
        self.setup_step(
            transfer.DirectoryUpload(workersrc="srcdir", masterdest=self.destdir))

        self.expect_commands(
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .exit(1))

        self.expect_outcome(result=FAILURE,
                           state_string="uploading srcdir (failure)")
        yield self.run_step()

    @defer.inlineCallbacks
    def testException(self):
        self.setup_step(
            transfer.DirectoryUpload(workersrc='srcdir', masterdest=self.destdir))

        writers = []

        self.expect_commands(
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"}, error=RuntimeError('uh oh'),
                             out_writers=writers))

        self.expect_outcome(
            result=EXCEPTION,
            state_string="uploading srcdir (exception)")
        yield self.run_step()

        self.assertEqual(len(writers), 1)
        self.assertEqual(writers[0].cancel.called, True)

        self.assertEqual(
            len(self.flushLoggedErrors(RuntimeError)), 1)

    def test_init_workersrc_keyword(self):
        step = transfer.DirectoryUpload(
            workersrc='srcfile', masterdest='dstfile')

        self.assertEqual(step.workersrc, 'srcfile')

    def test_init_workersrc_positional(self):
        step = transfer.DirectoryUpload('srcfile', 'dstfile')

        self.assertEqual(step.workersrc, 'srcfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.DirectoryUpload()
        with self.assertRaises(TypeError):
            transfer.DirectoryUpload('src')


class TestMultipleFileUpload(TestBuildStepMixin, TestReactorMixin,
                             unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.destdir = os.path.abspath('destdir')
        if os.path.exists(self.destdir):
            shutil.rmtree(self.destdir)

        return self.setup_test_build_step()

    def tearDown(self):
        if os.path.exists(self.destdir):
            shutil.rmtree(self.destdir)

        return self.tear_down_test_build_step()

    def testEmpty(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=[], masterdest=self.destdir))

        self.expect_commands()

        self.expect_outcome(result=SKIPPED, state_string="finished (skipped)")
        d = self.run_step()
        return d

    def testFile(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=["srcfile"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    def testDirectory(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=["srcdir"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file="srcdir", workdir='wkdir')
            .stat_dir()
            .exit(0),
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def test_not_existing_path(self):
        self.setup_step(transfer.MultipleFileUpload(workersrcs=["srcdir"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file='srcdir', workdir='wkdir')
            .exit(1))

        self.expect_outcome(result=FAILURE, state_string="uploading 1 file (failure)")
        self.expect_log_file('stderr',
                           "File wkdir/srcdir not available at worker")

        yield self.run_step()

    @defer.inlineCallbacks
    def test_special_path(self):
        self.setup_step(transfer.MultipleFileUpload(workersrcs=["srcdir"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file='srcdir', workdir='wkdir')
            .stat(mode=0)
            .exit(0))

        self.expect_outcome(result=FAILURE, state_string="uploading 1 file (failure)")
        self.expect_log_file('stderr', 'srcdir is neither a regular file, nor a directory')

        yield self.run_step()

    def testMultiple(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=["srcfile", "srcdir"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0),
            ExpectStat(file="srcdir", workdir='wkdir')
            .stat_dir()
            .exit(0),
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="uploading 2 files")
        d = self.run_step()
        return d

    def testMultipleString(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs="srcfile", masterdest=self.destdir))
        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))
        self.expect_outcome(
            result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    def testGlob(self):
        self.setup_step(
            transfer.MultipleFileUpload(
                workersrcs=["src*"], masterdest=self.destdir, glob=True))
        self.expect_commands(
            ExpectGlob(path=os.path.join('wkdir', 'src*'), log_environ=False)
            .files(["srcfile"])
            .exit(0),
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0)
        )
        self.expect_outcome(
            result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    def testFailedGlob(self):
        self.setup_step(
            transfer.MultipleFileUpload(
                workersrcs=["src*"], masterdest=self.destdir, glob=True))
        self.expect_commands(
            ExpectGlob(path=os.path.join('wkdir', 'src*'), log_environ=False)
            .files()
            .exit(1)
        )
        self.expect_outcome(
            result=SKIPPED, state_string="uploading 0 files (skipped)")
        d = self.run_step()
        return d

    def testFileWorker2_16(self):
        self.setup_step(
            transfer.MultipleFileUpload(
                workersrcs=["srcfile"], masterdest=self.destdir),
            worker_version={'*': '2.16'})

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(slavesrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    def testDirectoryWorker2_16(self):
        self.setup_step(
            transfer.MultipleFileUpload(
                workersrcs=["srcdir"], masterdest=self.destdir),
            worker_version={'*': '2.16'})

        self.expect_commands(
            ExpectStat(file="srcdir", workdir='wkdir')
            .stat_dir()
            .exit(0),
            ExpectUploadDirectory(slavesrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(result=SUCCESS, state_string="uploading 1 file")
        d = self.run_step()
        return d

    def testMultipleWorker2_16(self):
        self.setup_step(
            transfer.MultipleFileUpload(
                workersrcs=["srcfile", "srcdir"], masterdest=self.destdir),
            worker_version={'*': '2.16'})

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(slavesrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0),
            ExpectStat(file="srcdir", workdir='wkdir')
            .stat_dir()
            .exit(0),
            ExpectUploadDirectory(slavesrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="uploading 2 files")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def test_url(self):
        self.setup_step(transfer.MultipleFileUpload(workersrcs=["srcfile"], masterdest=self.destdir,
                                                   url="http://server/dir"))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectStat(file='srcfile', workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc='srcfile', workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading 1 file")

        yield self.run_step()

        self.step.addURL.assert_called_once_with("destdir", "http://server/dir")

    @defer.inlineCallbacks
    def test_url_text(self):
        self.setup_step(transfer.MultipleFileUpload(workersrcs=["srcfile"], masterdest=self.destdir,
                                                   url="http://server/dir", urlText='url text'))

        self.step.addURL = Mock()

        self.expect_commands(
            ExpectStat(file='srcfile', workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc='srcfile', workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0))

        self.expect_outcome(result=SUCCESS,
                           state_string="uploading 1 file")

        yield self.run_step()

        self.step.addURL.assert_called_once_with("url text", "http://server/dir")

    def testFailure(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=["srcfile", "srcdir"], masterdest=self.destdir))

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .exit(1))

        self.expect_outcome(
            result=FAILURE, state_string="uploading 2 files (failure)")
        d = self.run_step()
        return d

    @defer.inlineCallbacks
    def testException(self):
        self.setup_step(
            transfer.MultipleFileUpload(workersrcs=["srcfile", "srcdir"], masterdest=self.destdir))

        writers = []

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n", out_writers=writers, error=RuntimeError('uh oh')))

        self.expect_outcome(
            result=EXCEPTION, state_string="uploading 2 files (exception)")
        yield self.run_step()

        self.assertEqual(len(writers), 1)
        self.assertEqual(writers[0].cancel.called, True)

        self.assertEqual(
            len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def testSubclass(self):
        class CustomStep(transfer.MultipleFileUpload):
            uploadDone = Mock(return_value=None)
            allUploadsDone = Mock(return_value=None)

        step = CustomStep(
            workersrcs=["srcfile", "srcdir"], masterdest=self.destdir)
        self.setup_step(step)

        self.expect_commands(
            ExpectStat(file="srcfile", workdir='wkdir')
            .stat_file()
            .exit(0),
            ExpectUploadFile(workersrc="srcfile", workdir='wkdir',
                             blocksize=16384, maxsize=None, keepstamp=False,
                             writer=ExpectRemoteRef(remotetransfer.FileWriter))
            .upload_string("Hello world!\n")
            .exit(0),
            ExpectStat(file="srcdir", workdir='wkdir')
            .stat_dir()
            .exit(0),
            ExpectUploadDirectory(workersrc="srcdir", workdir='wkdir',
                                  blocksize=16384, compress=None, maxsize=None,
                                  writer=ExpectRemoteRef(remotetransfer.DirectoryWriter))
            .upload_tar_file('fake.tar', {"test": "Hello world!"})
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="uploading 2 files")

        yield self.run_step()

        self.assertEqual(step.uploadDone.call_count, 2)
        self.assertEqual(step.uploadDone.call_args_list[0],
                         ((SUCCESS, 'srcfile', os.path.join(self.destdir, 'srcfile')), {}))
        self.assertEqual(step.uploadDone.call_args_list[1],
                         ((SUCCESS, 'srcdir', os.path.join(self.destdir, 'srcdir')), {}))
        self.assertEqual(step.allUploadsDone.call_count, 1)
        self.assertEqual(step.allUploadsDone.call_args_list[0],
                         ((SUCCESS, ['srcfile', 'srcdir'], self.destdir), {}))

    def test_init_workersrcs_keyword(self):
        step = transfer.MultipleFileUpload(
            workersrcs=['srcfile'], masterdest='dstfile')

        self.assertEqual(step.workersrcs, ['srcfile'])

    def test_init_workersrcs_positional(self):
        step = transfer.MultipleFileUpload(['srcfile'], 'dstfile')

        self.assertEqual(step.workersrcs, ['srcfile'])

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.MultipleFileUpload()
        with self.assertRaises(TypeError):
            transfer.MultipleFileUpload(['srcfile'])


class TestFileDownload(TestBuildStepMixin, TestReactorMixin,
                       unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        fd, self.destfile = tempfile.mkstemp()
        os.close(fd)
        os.unlink(self.destfile)
        return self.setup_test_build_step()

    def tearDown(self):
        if os.path.exists(self.destfile):
            os.unlink(self.destfile)
        return self.tear_down_test_build_step()

    def test_init_workerdest_keyword(self):
        step = transfer.FileDownload(
            mastersrc='srcfile', workerdest='dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_workerdest_positional(self):
        step = transfer.FileDownload('srcfile', 'dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.FileDownload()
        with self.assertRaises(TypeError):
            transfer.FileDownload('srcfile')

    @defer.inlineCallbacks
    def testBasic(self):
        master_file = __file__
        self.setup_step(
            transfer.FileDownload(
                mastersrc=master_file, workerdest=self.destfile))

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(workerdest=self.destfile, workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.FileReader))
            .download_string(read.append, size=1000)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string=f"downloading to {os.path.basename(self.destfile)}")
        yield self.run_step()

        with open(master_file, "rb") as f:
            contents = f.read()
        contents = contents[:1000]
        self.assertEqual(b''.join(read), contents)

    @defer.inlineCallbacks
    def testBasicWorker2_16(self):
        master_file = __file__
        self.setup_step(
            transfer.FileDownload(
                mastersrc=master_file, workerdest=self.destfile),
            worker_version={'*': '2.16'})

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(slavedest=self.destfile, workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.FileReader))
            .download_string(read.append, size=1000)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS,
            state_string=f"downloading to {os.path.basename(self.destfile)}")
        yield self.run_step()

        with open(master_file, "rb") as f:
            contents = f.read()
        contents = contents[:1000]
        self.assertEqual(b''.join(read), contents)

    @defer.inlineCallbacks
    def test_no_file(self):
        self.setup_step(transfer.FileDownload(mastersrc='not existing file',
                                             workerdest=self.destfile))

        self.expect_commands()

        self.expect_outcome(result=FAILURE,
                           state_string=f"downloading to {os.path.basename(self.destfile)} "
                           "(failure)")
        self.expect_log_file('stderr',
                           "File 'not existing file' not available at master")
        yield self.run_step()


class TestStringDownload(TestBuildStepMixin, TestReactorMixin,
                         unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def tearDown(self):
        return self.tear_down_test_build_step()

    # check that ConfigErrors is raised on invalid 'mode' argument

    def testModeConfError(self):
        with self.assertRaisesRegex(config.ConfigErrors,
                                    "StringDownload step's mode must be an integer or None,"
                                    " got 'not-a-number'"):
            transfer.StringDownload("string", "file", mode="not-a-number")

    @defer.inlineCallbacks
    def testBasic(self):
        self.setup_step(transfer.StringDownload("Hello World", "hello.txt"))

        self.step.worker = Mock()
        self.step.remote = Mock()

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(workerdest="hello.txt", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .download_string(read.append)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="downloading to hello.txt")
        yield self.run_step()

        self.assertEqual(b''.join(read), b"Hello World")

    @defer.inlineCallbacks
    def testBasicWorker2_16(self):
        self.setup_step(
            transfer.StringDownload("Hello World", "hello.txt"),
            worker_version={'*': '2.16'})

        self.step.worker = Mock()
        self.step.remote = Mock()

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(slavedest="hello.txt", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .download_string(read.append)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="downloading to hello.txt")
        yield self.run_step()

        self.assertEqual(b''.join(read), b"Hello World")

    def testFailure(self):
        self.setup_step(transfer.StringDownload("Hello World", "hello.txt"))

        self.expect_commands(
            ExpectDownloadFile(workerdest="hello.txt", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .exit(1))

        self.expect_outcome(
            result=FAILURE, state_string="downloading to hello.txt (failure)")
        return self.run_step()

    def test_init_workerdest_keyword(self):
        step = transfer.StringDownload('srcfile', workerdest='dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_workerdest_positional(self):
        step = transfer.StringDownload('srcfile', 'dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.StringDownload()
        with self.assertRaises(TypeError):
            transfer.StringDownload('srcfile')


class TestJSONStringDownload(TestBuildStepMixin, TestReactorMixin,
                             unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def tearDown(self):
        return self.tear_down_test_build_step()

    @defer.inlineCallbacks
    def testBasic(self):
        msg = dict(message="Hello World")
        self.setup_step(transfer.JSONStringDownload(msg, "hello.json"))

        self.step.worker = Mock()
        self.step.remote = Mock()

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(workerdest="hello.json", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .download_string(read.append)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="downloading to hello.json")
        yield self.run_step()

        self.assertEqual(b''.join(read), b'{"message": "Hello World"}')

    @defer.inlineCallbacks
    def test_basic_with_renderable(self):
        msg = dict(message=Interpolate("Hello World"))
        self.setup_step(transfer.JSONStringDownload(msg, "hello.json"))

        self.step.worker = Mock()
        self.step.remote = Mock()

        # A place to store what gets read
        read = []

        self.expect_commands(
            ExpectDownloadFile(workerdest="hello.json", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .download_string(read.append)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="downloading to hello.json")
        yield self.run_step()

        self.assertEqual(b''.join(read), b'{"message": "Hello World"}')

    def testFailure(self):
        msg = dict(message="Hello World")
        self.setup_step(transfer.JSONStringDownload(msg, "hello.json"))

        self.expect_commands(
            ExpectDownloadFile(workerdest="hello.json", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .exit(1))

        self.expect_outcome(
            result=FAILURE, state_string="downloading to hello.json (failure)")
        return self.run_step()

    def test_init_workerdest_keyword(self):
        step = transfer.JSONStringDownload('srcfile', workerdest='dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_workerdest_positional(self):
        step = transfer.JSONStringDownload('srcfile', 'dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.JSONStringDownload()
        with self.assertRaises(TypeError):
            transfer.JSONStringDownload('srcfile')


class TestJSONPropertiesDownload(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def tearDown(self):
        return self.tear_down_test_build_step()

    @defer.inlineCallbacks
    def testBasic(self):
        self.setup_step(transfer.JSONPropertiesDownload("props.json"))
        self.step.build.setProperty('key1', 'value1', 'test')
        read = []
        self.expect_commands(
            ExpectDownloadFile(workerdest="props.json", workdir='wkdir',
                               blocksize=16384, maxsize=None, mode=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader))
            .download_string(read.append)
            .exit(0))

        self.expect_outcome(
            result=SUCCESS, state_string="downloading to props.json")
        yield self.run_step()
        # we decode as key order is dependent of python version
        self.assertEqual(json.loads((b''.join(read)).decode()), {
                         "properties": {"key1": "value1"}, "sourcestamps": []})

    def test_init_workerdest_keyword(self):
        step = transfer.JSONPropertiesDownload(workerdest='dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_workerdest_positional(self):
        step = transfer.JSONPropertiesDownload('dstfile')

        self.assertEqual(step.workerdest, 'dstfile')

    def test_init_positional_args(self):
        with self.assertRaises(TypeError):
            transfer.JSONPropertiesDownload()



# coding: utf-8
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import absolute_import
from __future__ import print_function

from twisted.trial import unittest

from buildbot_worker import util


class remove_userpassword(unittest.TestCase):

    def assertUrl(self, real_url, expected_url):
        new_url = util.remove_userpassword(real_url)
        self.assertEqual(expected_url, new_url)

    def test_url_with_no_user_and_password(self):
        self.assertUrl('http://myurl.com/myrepo', 'http://myurl.com/myrepo')

    def test_url_with_user_and_password(self):
        self.assertUrl(
            'http://myuser:mypass@myurl.com/myrepo', 'http://myurl.com/myrepo')

    def test_another_url_with_no_user_and_password(self):
        self.assertUrl(
            'http://myurl2.com/myrepo2', 'http://myurl2.com/myrepo2')

    def test_another_url_with_user_and_password(self):
        self.assertUrl(
            'http://myuser2:mypass2@myurl2.com/myrepo2', 'http://myurl2.com/myrepo2')

    def test_with_different_protocol_without_user_and_password(self):
        self.assertUrl('ssh://myurl3.com/myrepo3', 'ssh://myurl3.com/myrepo3')

    def test_with_different_protocol_with_user_and_password(self):
        self.assertUrl(
            'ssh://myuser3:mypass3@myurl3.com/myrepo3', 'ssh://myurl3.com/myrepo3')

    def test_file_path(self):
        self.assertUrl('/home/me/repos/my-repo', '/home/me/repos/my-repo')

    def test_file_path_with_at_sign(self):
        self.assertUrl('/var/repos/speci@l', '/var/repos/speci@l')

    def test_win32file_path(self):
        self.assertUrl('c:\\repos\\my-repo', 'c:\\repos\\my-repo')


class TestObfuscated(unittest.TestCase):

    def testSimple(self):
        c = util.Obfuscated('real', '****')
        self.assertEqual(str(c), '****')
        self.assertEqual(repr(c), "'****'")

    def testObfuscatedCommand(self):
        cmd = ['echo', util.Obfuscated('password', '*******')]
        cmd_bytes = [b'echo', util.Obfuscated(b'password', b'*******')]
        cmd_unicode = [u'echo', util.Obfuscated(u'password', u'привет')]

        self.assertEqual(
            ['echo', 'password'], util.Obfuscated.get_real(cmd))
        self.assertEqual(
            ['echo', '*******'], util.Obfuscated.get_fake(cmd))
        self.assertEqual(
            [b'echo', b'password'], util.Obfuscated.get_real(cmd_bytes))
        self.assertEqual(
            [b'echo', b'*******'], util.Obfuscated.get_fake(cmd_bytes))
        self.assertEqual(
            [u'echo', u'password'], util.Obfuscated.get_real(cmd_unicode))
        self.assertEqual(
            [u'echo', u'привет'], util.Obfuscated.get_fake(cmd_unicode))

    def testObfuscatedNonString(self):
        cmd = ['echo', 1]
        cmd_bytes = [b'echo', 2]
        cmd_unicode = [u'привет', 3]
        self.assertEqual(['echo', '1'], util.Obfuscated.get_real(cmd))
        self.assertEqual([b'echo', '2'], util.Obfuscated.get_fake(cmd_bytes))
        self.assertEqual([u'привет', u'3'],
                         util.Obfuscated.get_fake(cmd_unicode))

    def testObfuscatedNonList(self):
        cmd = 1
        self.assertEqual(1, util.Obfuscated.get_real(cmd))
        self.assertEqual(1, util.Obfuscated.get_fake(cmd))


class TestRewrap(unittest.TestCase):

    def test_main(self):
        tests = [
            ("", "", None),
            ("\n", "\n", None),
            ("\n  ", "\n", None),
            ("  \n", "\n", None),
            ("  \n  ", "\n", None),
            ("""
                multiline
                with
                indent
                """,
             "\nmultiline with indent",
             None),
            ("""\
                multiline
                with
                indent

                """,
             "multiline with indent\n",
             None),
            ("""\
                 multiline
                 with
                 indent

                 """,
             "multiline with indent\n",
             None),
            ("""\
                multiline
                with
                indent

                  and
                   formatting
                """,
             "multiline with indent\n  and\n   formatting\n",
             None),
            ("""\
                multiline
                with
                indent
                and wrapping

                  and
                   formatting
                """,
             "multiline with\nindent and\nwrapping\n  and\n   formatting\n",
             15),
        ]

        for text, expected, width in tests:
            self.assertEqual(util.rewrap(text, width=width), expected)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.changes import base
from buildbot.config import ConfigErrors
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import changesource
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class TestChangeSource(changesource.ChangeSourceMixin,
                       TestReactorMixin,
                       unittest.TestCase):
    timeout = 120

    class Subclass(base.ChangeSource):
        pass

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpChangeSource()

    def tearDown(self):
        return self.tearDownChangeSource()

    @defer.inlineCallbacks
    def test_activation(self):
        cs = self.Subclass(name="DummyCS")
        cs.activate = mock.Mock(return_value=defer.succeed(None))
        cs.deactivate = mock.Mock(return_value=defer.succeed(None))

        # set the changesourceid, and claim the changesource on another master
        yield self.attachChangeSource(cs)
        self.setChangeSourceToMaster(self.OTHER_MASTER_ID)

        yield cs.startService()
        cs.clock.advance(cs.POLL_INTERVAL_SEC / 2)
        cs.clock.advance(cs.POLL_INTERVAL_SEC / 5)
        cs.clock.advance(cs.POLL_INTERVAL_SEC / 5)
        self.assertFalse(cs.activate.called)
        self.assertFalse(cs.deactivate.called)
        self.assertFalse(cs.active)
        self.assertEqual(cs.serviceid, self.DUMMY_CHANGESOURCE_ID)

        # clear that masterid
        yield cs.stopService()
        self.setChangeSourceToMaster(None)

        yield cs.startService()
        cs.clock.advance(cs.POLL_INTERVAL_SEC)
        self.assertTrue(cs.activate.called)
        self.assertFalse(cs.deactivate.called)
        self.assertTrue(cs.active)

        # stop the service and see that deactivate is called
        yield cs.stopService()
        self.assertTrue(cs.activate.called)
        self.assertTrue(cs.deactivate.called)
        self.assertFalse(cs.active)


class TestPollingChangeSource(changesource.ChangeSourceMixin,
                              TestReactorMixin,
                              unittest.TestCase):
    timeout = 120

    class Subclass(base.PollingChangeSource):
        pass

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpChangeSource()

        with assertProducesWarnings(DeprecatedApiWarning,
                                    message_pattern="use ReconfigurablePollingChangeSource"):
            cs = self.Subclass(name="DummyCS")
        yield self.attachChangeSource(cs)

    def tearDown(self):
        return self.tearDownChangeSource()

    @defer.inlineCallbacks
    def runClockFor(self, _, secs):
        yield self.reactor.pump([0] + [1.0] * secs)

    def test_loop_loops(self):
        # track when poll() gets called
        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())

        self.changesource.pollInterval = 5
        self.startChangeSource()
        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)

        @d.addCallback
        def check(_):
            # note that it does *not* poll at time 0
            self.assertEqual(loops, [5.0, 10.0])
        self.reactor.callWhenRunning(d.callback, None)
        return d

    def test_loop_exception(self):
        # track when poll() gets called
        loops = []

        def poll():
            loops.append(self.reactor.seconds())
            raise RuntimeError("oh noes")
        self.changesource.poll = poll

        self.changesource.pollInterval = 5
        self.startChangeSource()

        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)

        @d.addCallback
        def check(_):
            # note that it keeps looping after error
            self.assertEqual(loops, [5.0, 10.0])
            self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 2)
        self.reactor.callWhenRunning(d.callback, None)
        return d

    def test_poll_only_if_activated(self):
        """The polling logic only applies if the source actually starts!"""

        self.setChangeSourceToMaster(self.OTHER_MASTER_ID)

        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())

        self.changesource.pollInterval = 5
        self.startChangeSource()

        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)

        @d.addCallback
        def check(_):
            # it doesn't do anything because it was already claimed
            self.assertEqual(loops, [])

        self.reactor.callWhenRunning(d.callback, None)
        return d

    def test_pollAtLaunch(self):
        # track when poll() gets called
        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())

        self.changesource.pollInterval = 5
        self.changesource.pollAtLaunch = True
        self.startChangeSource()

        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)

        @d.addCallback
        def check(_):
            # note that it *does* poll at time 0
            self.assertEqual(loops, [0.0, 5.0, 10.0])
        self.reactor.callWhenRunning(d.callback, None)
        return d


class TestReconfigurablePollingChangeSource(changesource.ChangeSourceMixin,
                                            TestReactorMixin,
                                            unittest.TestCase):

    class Subclass(base.ReconfigurablePollingChangeSource):
        pass

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        yield self.setUpChangeSource()

        yield self.attachChangeSource(self.Subclass(name="DummyCS"))

    def tearDown(self):
        return self.tearDownChangeSource()

    @defer.inlineCallbacks
    def runClockFor(self, secs):
        yield self.reactor.pump([0] + [1.0] * secs)

    @defer.inlineCallbacks
    def test_config_negative_interval(self):
        try:
            yield self.changesource.reconfigServiceWithSibling(self.Subclass(
                name="NegativePollInterval", pollInterval=-1, pollAtLaunch=False))
        except ConfigErrors as e:
            self.assertEqual("interval must be >= 0: -1", e.errors[0])

    @defer.inlineCallbacks
    def test_config_negative_random_delay_min(self):
        try:
            yield self.changesource.reconfigServiceWithSibling(
                self.Subclass(name="NegativePollRandomDelayMin", pollInterval=1,
                              pollAtLaunch=False, pollRandomDelayMin=-1, pollRandomDelayMax=1))
        except ConfigErrors as e:
            self.assertEqual("min random delay must be >= 0: -1", e.errors[0])

    @defer.inlineCallbacks
    def test_config_negative_random_delay_max(self):
        try:
            yield self.changesource.reconfigServiceWithSibling(
                self.Subclass(name="NegativePollRandomDelayMax", pollInterval=1,
                              pollAtLaunch=False, pollRandomDelayMin=1, pollRandomDelayMax=-1))
        except ConfigErrors as e:
            self.assertEqual("max random delay must be >= 0: -1", e.errors[0])

    @defer.inlineCallbacks
    def test_config_random_delay_min_gt_random_delay_max(self):
        try:
            yield self.changesource.reconfigServiceWithSibling(
                self.Subclass(name="PollRandomDelayMinGtPollRandomDelayMax", pollInterval=1,
                              pollAtLaunch=False, pollRandomDelayMin=2, pollRandomDelayMax=1))
        except ConfigErrors as e:
            self.assertEqual("min random delay must be <= 1: 2", e.errors[0])

    @defer.inlineCallbacks
    def test_config_random_delay_max_gte_interval(self):
        try:
            yield self.changesource.reconfigServiceWithSibling(
                self.Subclass(name="PollRandomDelayMaxGtePollInterval", pollInterval=1,
                              pollAtLaunch=False, pollRandomDelayMax=1))
        except ConfigErrors as e:
            self.assertEqual("max random delay must be < 1: 1", e.errors[0])

    @defer.inlineCallbacks
    def test_loop_loops(self):
        # track when poll() gets called
        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())

        yield self.startChangeSource()
        yield self.changesource.reconfigServiceWithSibling(self.Subclass(
            name="DummyCS", pollInterval=5, pollAtLaunch=False))

        yield self.runClockFor(12)
        # note that it does *not* poll at time 0
        self.assertEqual(loops, [5.0, 10.0])

    @defer.inlineCallbacks
    def test_loop_exception(self):
        # track when poll() gets called
        loops = []

        def poll():
            loops.append(self.reactor.seconds())
            raise RuntimeError("oh noes")
        self.changesource.poll = poll

        yield self.startChangeSource()
        yield self.changesource.reconfigServiceWithSibling(self.Subclass(
            name="DummyCS", pollInterval=5, pollAtLaunch=False))

        yield self.runClockFor(12)
        # note that it keeps looping after error
        self.assertEqual(loops, [5.0, 10.0])
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 2)

    @defer.inlineCallbacks
    def test_poll_only_if_activated(self):
        """The polling logic only applies if the source actually starts!"""

        self.setChangeSourceToMaster(self.OTHER_MASTER_ID)

        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())

        yield self.startChangeSource()
        yield self.changesource.reconfigServiceWithSibling(self.Subclass(
            name="DummyCS", pollInterval=5, pollAtLaunch=False))

        yield self.runClockFor(12)

        # it doesn't do anything because it was already claimed
        self.assertEqual(loops, [])

    @defer.inlineCallbacks
    def test_pollAtLaunch(self):
        # track when poll() gets called
        loops = []
        self.changesource.poll = \
            lambda: loops.append(self.reactor.seconds())
        yield self.startChangeSource()
        yield self.changesource.reconfigServiceWithSibling(self.Subclass(
            name="DummyCS", pollInterval=5, pollAtLaunch=True))

        yield self.runClockFor(12)

        # note that it *does* poll at time 0
        self.assertEqual(loops, [0.0, 5.0, 10.0])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import shutil
import sys
from subprocess import call
from subprocess import check_call
from textwrap import dedent

from twisted.trial import unittest


class BuildbotWWWPkg(unittest.TestCase):
    pkgName = "buildbot_www"
    pkgPaths = ["www", "base"]
    epName = "base"

    loadTestScript = dedent("""
        import pkg_resources
        import re
        apps = {}
        for ep in pkg_resources.iter_entry_points('buildbot.www'):
            apps[ep.name] = ep.load()

        print(apps["%(epName)s"])
        assert("scripts.js" in apps["%(epName)s"].resource.listNames())
        assert(re.match(r'\d+\.\d+\.\d+', apps["%(epName)s"].version) is not None)
        assert(apps["%(epName)s"].description is not None)
        """)

    @property
    def path(self):
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *self.pkgPaths))

    def rmtree(self, d):
        if os.path.isdir(d):
            shutil.rmtree(d)

    def setUp(self):
        call("pip uninstall -y " + self.pkgName, shell=True)
        self.rmtree(os.path.join(self.path, "build"))
        self.rmtree(os.path.join(self.path, "dist"))
        self.rmtree(os.path.join(self.path, "static"))

    def run_setup(self, cmd):
        check_call([sys.executable, 'setup.py', cmd], cwd=self.path)

    def check_correct_installation(self):
        # assert we can import buildbot_www
        # and that it has an endpoint with resource containing file "script.js"
        check_call([
            sys.executable, '-c', self.loadTestScript % dict(epName=self.epName)])

    def test_install(self):
        self.run_setup("install")
        self.check_correct_installation()

    def test_wheel(self):
        self.run_setup("bdist_wheel")
        check_call("pip install dist/*.whl", shell=True, cwd=self.path)
        self.check_correct_installation()

    def test_develop(self):
        self.run_setup("develop")
        self.check_correct_installation()

    def test_develop_via_pip(self):
        check_call("pip install -e .", shell=True, cwd=self.path)
        self.check_correct_installation()

    def test_sdist(self):
        self.run_setup("sdist")
        check_call("pip install dist/*.tar.gz", shell=True, cwd=self.path)
        self.check_correct_installation()


class BuildbotConsolePkg(BuildbotWWWPkg):
    pkgName = "buildbot-console-view"
    pkgPaths = ["www", "console_view"]
    epName = "console_view"


class BuildbotWaterfallPkg(BuildbotWWWPkg):
    pkgName = "buildbot-waterfall-view"
    pkgPaths = ["www", "waterfall_view"]
    epName = "waterfall_view"


class BuildbotCodeparameterPkg(BuildbotWWWPkg):
    pkgName = "buildbot-codeparameter"
    pkgPaths = ["www", "codeparameter"]
    epName = "codeparameter"



# coding: utf-8
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import types

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.www import WwwTestMixin
from buildbot.www import avatar
from buildbot.www import ldapuserinfo

try:
    import ldap3
except ImportError:
    ldap3 = None


def get_config_parameter(p):
    params = {'DEFAULT_SERVER_ENCODING': 'utf-8'}
    return params[p]


fake_ldap = types.ModuleType('ldap3')
fake_ldap.SEARCH_SCOPE_WHOLE_SUBTREE = 2
fake_ldap.get_config_parameter = get_config_parameter


class FakeLdap:

    def __init__(self):
        def search(base, filterstr='f', scope=None, attributes=None):
            pass
        self.search = mock.Mock(spec=search)


class CommonTestCase(unittest.TestCase):
    """Common fixture for all ldapuserinfo tests

    we completely fake the ldap3 module, so no need to require
    it to run the unit tests
    """

    if not ldap3:
        skip = 'ldap3 is required for LdapUserInfo tests'

    def setUp(self):
        self.ldap = FakeLdap()
        self.makeUserInfoProvider()
        self.userInfoProvider.connectLdap = lambda: self.ldap

        def search(base, filterstr='f', attributes=None):
            pass
        self.userInfoProvider.search = mock.Mock(spec=search)

    def makeUserInfoProvider(self):
        """To be implemented by subclasses"""
        raise NotImplementedError

    def _makeSearchSideEffect(self, attribute_type, ret):
        ret = [[{'dn': i[0], attribute_type: i[1]} for i in r]
               for r in ret]
        self.userInfoProvider.search.side_effect = ret

    def makeSearchSideEffect(self, ret):
        return self._makeSearchSideEffect('attributes', ret)

    def makeRawSearchSideEffect(self, ret):
        return self._makeSearchSideEffect('raw_attributes', ret)

    def assertSearchCalledWith(self, exp):
        got = self.userInfoProvider.search.call_args_list
        self.assertEqual(len(exp), len(got))
        for i, val in enumerate(exp):
            self.assertEqual(val[0][0], got[i][0][1])
            self.assertEqual(val[0][1], got[i][0][2])
            self.assertEqual(val[0][2], got[i][1]['attributes'])


class LdapUserInfo(CommonTestCase):

    def makeUserInfoProvider(self):
        self.userInfoProvider = ldapuserinfo.LdapUserInfo(
            uri="ldap://uri", bindUser="user", bindPw="pass",
            accountBase="accbase", groupBase="groupbase",
            accountPattern="accpattern", groupMemberPattern="groupMemberPattern",
            accountFullName="accountFullName",
            accountEmail="accountEmail",
            groupName="groupName",
            avatarPattern="avatar",
            avatarData="picture",
            accountExtraFields=["myfield"])

    @defer.inlineCallbacks
    def test_updateUserInfoNoResults(self):
        self.makeSearchSideEffect([[], [], []])

        try:
            yield self.userInfoProvider.getUserInfo("me")
        except KeyError as e:
            self.assertRegex(
                repr(e), r"KeyError\('ldap search \"accpattern\" returned 0 results',?\)")
        else:
            self.fail("should have raised a key error")

    @defer.inlineCallbacks
    def test_updateUserInfoNoGroups(self):
        self.makeSearchSideEffect([[(
            "cn", {"accountFullName": "me too",
                   "accountEmail": "mee@too"})], [], []])
        res = yield self.userInfoProvider.getUserInfo("me")
        self.assertSearchCalledWith([
            (('accbase', 'accpattern',
              ['accountEmail', 'accountFullName', 'myfield']), {}),
            (('groupbase', 'groupMemberPattern', ['groupName']), {}),
        ])
        self.assertEqual(res, {'email': 'mee@too', 'full_name': 'me too',
                               'groups': [], 'username': 'me'})

    @defer.inlineCallbacks
    def test_updateUserInfoGroups(self):
        self.makeSearchSideEffect([[("cn", {"accountFullName": "me too",
                                            "accountEmail": "mee@too"})],
                                   [("cn", {"groupName": ["group"]}),
                                    ("cn", {"groupName": ["group2"]})
                                    ], []])
        res = yield self.userInfoProvider.getUserInfo("me")
        self.assertEqual(res, {'email': 'mee@too', 'full_name': 'me too',
                               'groups': ["group", "group2"], 'username': 'me'})

    @defer.inlineCallbacks
    def test_updateUserInfoGroupsUnicodeDn(self):
        # In case of non Ascii DN, ldap3 lib returns an UTF-8 str
        dn = "cn=Sébastien,dc=example,dc=org"
        # If groupMemberPattern is an str, and dn is not decoded,
        # the resulting filter will be an str, leading to UnicodeDecodeError
        # in ldap3.protocol.convert.validate_assertion_value()
        # So we use an unicode pattern:
        self.userInfoProvider.groupMemberPattern = '(member=%(dn)s)'
        self.makeSearchSideEffect([[(dn, {"accountFullName": "me too",
                                          "accountEmail": "mee@too"})],
                                   [("cn", {"groupName": ["group"]}),
                                    ("cn", {"groupName": ["group2"]})
                                    ], []])
        res = yield self.userInfoProvider.getUserInfo("me")
        self.assertEqual(res, {'email': 'mee@too', 'full_name': 'me too',
                               'groups': ["group", "group2"], 'username': 'me'})


class LdapAvatar(CommonTestCase, TestReactorMixin, WwwTestMixin):
    @defer.inlineCallbacks
    def setUp(self):
        CommonTestCase.setUp(self)
        self.setup_test_reactor()

        master = self.make_master(
            url='http://a/b/',
            avatar_methods=[self.userInfoProvider])

        self.rsrc = avatar.AvatarResource(master)
        self.rsrc.reconfigResource(master.config)

        yield self.master.startService()

    def makeUserInfoProvider(self):
        self.userInfoProvider = ldapuserinfo.LdapUserInfo(
            uri="ldap://uri", bindUser="user", bindPw="pass",
            accountBase="accbase", groupBase="groupbase",
            accountPattern="accpattern=%(username)s", groupMemberPattern="groupMemberPattern",
            accountFullName="accountFullName",
            accountEmail="accountEmail",
            groupName="groupName",
            avatarPattern="avatar=%(email)s",
            avatarData="picture",
            accountExtraFields=["myfield"])

    @defer.inlineCallbacks
    def _getUserAvatar(self, mimeTypeAndData):
        _, data = mimeTypeAndData
        self.makeRawSearchSideEffect([
            [("cn", {"picture": [data]})]])
        res = yield self.render_resource(self.rsrc, b'/?email=me')
        self.assertSearchCalledWith([
            (('accbase', 'avatar=me', ['picture']), {}),
        ])
        return res

    @defer.inlineCallbacks
    def test_getUserAvatarPNG(self):
        mimeTypeAndData = (b'image/png', b'\x89PNG lljklj')
        yield self._getUserAvatar(mimeTypeAndData)
        self.assertRequest(contentType=mimeTypeAndData[0],
            content=mimeTypeAndData[1])

    @defer.inlineCallbacks
    def test_getUserAvatarJPEG(self):
        mimeTypeAndData = (b'image/jpeg', b'\xff\xd8\xff lljklj')
        yield self._getUserAvatar(mimeTypeAndData)
        self.assertRequest(contentType=mimeTypeAndData[0],
            content=mimeTypeAndData[1])

    @defer.inlineCallbacks
    def test_getUserAvatarGIF(self):
        mimeTypeAndData = (b'image/gif', b'GIF8 lljklj')
        yield self._getUserAvatar(mimeTypeAndData)
        self.assertRequest(contentType=mimeTypeAndData[0],
            content=mimeTypeAndData[1])

    @defer.inlineCallbacks
    def test_getUserAvatarUnknownType(self):
        mimeTypeAndData = (b'', b'unknown image format')
        res = yield self._getUserAvatar(mimeTypeAndData)
        # Unknown format means data won't be sent
        self.assertEqual(res, dict(redirected=b'img/nobody.png'))

    @defer.inlineCallbacks
    def test_getUsernameAvatar(self):
        mimeType = b'image/gif'
        data = b'GIF8 lljklj'
        self.makeRawSearchSideEffect([
            [("cn", {"picture": [data]})]])
        yield self.render_resource(self.rsrc, b'/?username=me')
        self.assertSearchCalledWith([
            (('accbase', 'accpattern=me', ['picture']), {}),
        ])
        self.assertRequest(contentType=mimeType,
            content=data)

    @defer.inlineCallbacks
    def test_getUnknownUsernameAvatar(self):
        self.makeSearchSideEffect([[], [], []])
        res = yield self.render_resource(self.rsrc, b'/?username=other')
        self.assertSearchCalledWith([
            (('accbase', 'accpattern=other', ['picture']), {}),
        ])
        self.assertEqual(res, dict(redirected=b'img/nobody.png'))


class LdapUserInfoNotEscCharsDn(CommonTestCase):
    def makeUserInfoProvider(self):
        self.userInfoProvider = ldapuserinfo.LdapUserInfo(
            uri="ldap://uri", bindUser="user", bindPw="pass",
            accountBase="accbase", groupBase="groupbase",
            accountPattern="accpattern", groupMemberPattern="(member=%(dn)s)",
            accountFullName="accountFullName",
            accountEmail="accountEmail",
            groupName="groupName",
            avatarPattern="avatar",
            avatarData="picture")

    @defer.inlineCallbacks
    def test_getUserInfoGroupsNotEscCharsDn(self):
        dn = "cn=Lastname, Firstname \28UIDxxx\29,dc=example,dc=org"
        pattern = self.userInfoProvider.groupMemberPattern % dict(dn=dn)
        self.makeSearchSideEffect([[(dn, {"accountFullName": "Lastname, Firstname (UIDxxx)",
                                          "accountEmail": "mee@too"})],
                                   [("cn", {"groupName": ["group"]}),
                                    ("cn", {"groupName": ["group2"]})
                                    ], []])
        res = yield self.userInfoProvider.getUserInfo("me")
        self.assertSearchCalledWith([
            (('accbase', 'accpattern',
              ['accountEmail', 'accountFullName']), {}),
            (('groupbase', pattern, ['groupName']), {}),
        ])
        self.assertEqual(res, {'email': 'mee@too',
                               'full_name': 'Lastname, Firstname (UIDxxx)',
                               'groups': ["group", "group2"],
                               'username': 'me'})


class LdapUserInfoNoGroups(CommonTestCase):

    def makeUserInfoProvider(self):
        self.userInfoProvider = ldapuserinfo.LdapUserInfo(
            uri="ldap://uri", bindUser="user", bindPw="pass",
            accountBase="accbase",
            accountPattern="accpattern",
            accountFullName="accountFullName",
            accountEmail="accountEmail",
            avatarPattern="avatar",
            avatarData="picture",
            accountExtraFields=["myfield"])

    @defer.inlineCallbacks
    def test_updateUserInfo(self):
        self.makeSearchSideEffect([[(
            "cn", {"accountFullName": "me too",
                   "accountEmail": "mee@too"})], [], []])
        res = yield self.userInfoProvider.getUserInfo("me")
        self.assertSearchCalledWith([
            (('accbase', 'accpattern',
              ['accountEmail', 'accountFullName', 'myfield']), {}),
        ])
        self.assertEqual(res, {'email': 'mee@too', 'full_name': 'me too',
                               'groups': [], 'username': 'me'})


class Config(unittest.TestCase):
    if not ldap3:
        skip = 'ldap3 is required for LdapUserInfo tests'

    def test_missing_group_name(self):
        with self.assertRaises(ValueError):
            ldapuserinfo.LdapUserInfo(groupMemberPattern="member=%(dn)s",
                                      groupBase="grpbase", uri="ldap://uri",
                                      bindUser="user", bindPw="pass",
                                      accountBase="accbase",
                                      accountPattern="accpattern",
                                      accountFullName="accountFullName",
                                      accountEmail="accountEmail")

    def test_missing_group_base(self):
        with self.assertRaises(ValueError):
            ldapuserinfo.LdapUserInfo(groupMemberPattern="member=%(dn)s",
                                      groupName="group",
                                      uri="ldap://uri", bindUser="user",
                                      bindPw="pass", accountBase="accbase",
                                      accountPattern="accpattern",
                                      accountFullName="accountFullName",
                                      accountEmail="accountEmail")

    def test_missing_two_params(self):
        with self.assertRaises(ValueError):
            ldapuserinfo.LdapUserInfo(groupName="group", uri="ldap://uri",
                                      bindUser="user", bindPw="pass",
                                      accountBase="accbase",
                                      accountPattern="accpattern",
                                      accountFullName="accountFullName",
                                      accountEmail="accountEmail")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Unit tests for the plugin framework
"""

import warnings

import mock

from twisted.trial import unittest
from zope.interface import implementer

import buildbot.plugins.db
from buildbot.errors import PluginDBError
from buildbot.interfaces import IPlugin
from buildbot.test.util.warnings import assertProducesWarning

# buildbot.plugins.db needs to be imported for patching, however just 'db' is
# much shorter for using in tests
db = buildbot.plugins.db


class FakeEntry:

    """
    An entry suitable for unit tests
    """

    def __init__(self, name, project_name, version, fail_require, value, warnings=None):
        self._name = name
        self._dist = mock.Mock(spec_set=['project_name', 'version'])
        self._dist.project_name = project_name
        self._dist.version = version
        self._fail_require = fail_require
        self._value = value
        self._warnings = [] if warnings is None else warnings

    @property
    def name(self):
        "entry name"
        return self._name

    @property
    def dist(self):
        "dist thingie"
        return self._dist

    def require(self):
        """
        handle external dependencies
        """
        if self._fail_require:
            raise RuntimeError('Fail require as requested')

    def load(self):
        """
        handle loading
        """
        for w in self._warnings:
            warnings.warn(w, DeprecationWarning)
        return self._value


class ITestInterface(IPlugin):

    """
    test interface
    """
    def hello(self):
        pass


@implementer(ITestInterface)
class ClassWithInterface:

    """
    a class to implement a simple interface
    """

    def __init__(self, name=None):
        self._name = name

    def hello(self, name=None):
        'implement the required method'
        return name or self._name


class ClassWithNoInterface:

    """
    just a class
    """


# NOTE: buildbot.plugins.db prepends the group with common namespace --
# 'buildbot.'
_FAKE_ENTRIES = {
    'buildbot.interface': [
        FakeEntry('good', 'non-existent', 'irrelevant', False,
                  ClassWithInterface),
        FakeEntry('deep.path', 'non-existent', 'irrelevant', False,
                  ClassWithInterface)
    ],
    'buildbot.interface_warnings': [
        FakeEntry('good', 'non-existent', 'irrelevant', False,
                  ClassWithInterface, warnings=['test warning']),
        FakeEntry('deep.path', 'non-existent', 'irrelevant', False,
                  ClassWithInterface, warnings=['test warning'])
    ],
    'buildbot.interface_failed': [
        FakeEntry('good', 'non-existent', 'irrelevant', True,
                  ClassWithInterface)
    ],
    'buildbot.no_interface': [
        FakeEntry('good', 'non-existent', 'irrelevant', False,
                  ClassWithNoInterface)
    ],
    'buildbot.no_interface_again': [
        FakeEntry('good', 'non-existent', 'irrelevant', False,
                  ClassWithNoInterface)
    ],
    'buildbot.no_interface_failed': [
        FakeEntry('good', 'non-existent', 'irrelevant', True,
                  ClassWithNoInterface)
    ],
    'buildbot.duplicates': [
        FakeEntry('good', 'non-existent', 'first', False,
                  ClassWithNoInterface),
        FakeEntry('good', 'non-existent', 'second', False,
                  ClassWithNoInterface)
    ]
}


def provide_fake_entries(group):
    """
    give a set of fake entries for known groups
    """
    return _FAKE_ENTRIES.get(group, [])


@mock.patch('buildbot.plugins.db.iter_entry_points', provide_fake_entries)
class TestBuildbotPlugins(unittest.TestCase):

    def setUp(self):
        buildbot.plugins.db._DB = buildbot.plugins.db._PluginDB()

    def test_check_group_registration(self):
        with mock.patch.object(buildbot.plugins.db, '_DB', db._PluginDB()):
            # The groups will be prepended with namespace, so info() will
            # return a dictionary with right keys, but no data
            groups = set(_FAKE_ENTRIES.keys())
            for group in groups:
                db.get_plugins(group)

            registered = set(db.info().keys())
            self.assertEqual(registered, groups)
            self.assertEqual(registered, set(db.namespaces()))

    def test_interface_provided_simple(self):
        # Basic check before the actual test
        self.assertTrue(ITestInterface.implementedBy(ClassWithInterface))

        plugins = db.get_plugins('interface', interface=ITestInterface)

        self.assertTrue('good' in plugins.names)

        result_get = plugins.get('good')
        result_getattr = plugins.good
        self.assertFalse(result_get is None)
        self.assertTrue(result_get is result_getattr)

        # Make sure we actually got our class
        greeter = result_get('yes')
        self.assertEqual('yes', greeter.hello())
        self.assertEqual('no', greeter.hello('no'))

    def test_missing_plugin(self):
        plugins = db.get_plugins('interface', interface=ITestInterface)

        with self.assertRaises(AttributeError):
            getattr(plugins, 'bad')
        with self.assertRaises(PluginDBError):
            plugins.get('bad')
        with self.assertRaises(PluginDBError):
            plugins.get('good.extra')

    def test_interface_provided_deep(self):
        # Basic check before the actual test
        self.assertTrue(ITestInterface.implementedBy(ClassWithInterface))

        plugins = db.get_plugins('interface', interface=ITestInterface)

        self.assertTrue('deep.path' in plugins.names)

        self.assertTrue('deep.path' in plugins)
        self.assertFalse('even.deeper.path' in plugins)

        result_get = plugins.get('deep.path')
        result_getattr = plugins.deep.path
        self.assertFalse(result_get is None)
        self.assertTrue(result_get is result_getattr)

        # Make sure we actually got our class
        greeter = result_get('yes')
        self.assertEqual('yes', greeter.hello())
        self.assertEqual('no', greeter.hello('no'))

    def test_interface_warnings(self):
        # we should not get no warnings when not trying to access the plugin
        plugins = db.get_plugins('interface_warnings', interface=ITestInterface)
        self.assertTrue('good' in plugins.names)
        self.assertTrue('deep.path' in plugins.names)

        # we should get warning when trying to access the plugin
        with assertProducesWarning(DeprecationWarning, "test warning"):
            _ = plugins.get('good')
        with assertProducesWarning(DeprecationWarning, "test warning"):
            _ = plugins.good
        with assertProducesWarning(DeprecationWarning, "test warning"):
            _ = plugins.get('deep.path')
        with assertProducesWarning(DeprecationWarning, "test warning"):
            _ = plugins.deep.path

    def test_interface_provided_deps_failed(self):
        plugins = db.get_plugins('interface_failed', interface=ITestInterface,
                                 check_extras=True)
        with self.assertRaises(PluginDBError):
            plugins.get('good')

    def test_required_interface_not_provided(self):
        plugins = db.get_plugins('no_interface_again',
                                 interface=ITestInterface)
        self.assertTrue(plugins._interface is ITestInterface)
        with self.assertRaises(PluginDBError):
            plugins.get('good')

    def test_no_interface_provided(self):
        plugins = db.get_plugins('no_interface')
        self.assertFalse(plugins.get('good') is None)

    def test_no_interface_provided_deps_failed(self):
        plugins = db.get_plugins('no_interface_failed', check_extras=True)
        with self.assertRaises(PluginDBError):
            plugins.get('good')

    def test_failure_on_dups(self):
        with self.assertRaises(PluginDBError):
            db.get_plugins('duplicates', load_now=True)

    def test_get_info_on_a_known_plugin(self):
        plugins = db.get_plugins('interface')
        self.assertEqual(('non-existent', 'irrelevant'), plugins.info('good'))

    def test_failure_on_unknown_plugin_info(self):
        plugins = db.get_plugins('interface')
        with self.assertRaises(PluginDBError):
            plugins.info('bad')

    def test_failure_on_unknown_plugin_get(self):
        plugins = db.get_plugins('interface')
        with self.assertRaises(PluginDBError):
            plugins.get('bad')


class SimpleFakeEntry(FakeEntry):

    def __init__(self, name, value):
        super().__init__(name, 'non-existent', 'irrelevant', False, value)


_WORKER_FAKE_ENTRIES = {
    'buildbot.worker': [
        SimpleFakeEntry('Worker', ClassWithInterface),
        SimpleFakeEntry('EC2LatentWorker', ClassWithInterface),
        SimpleFakeEntry('LibVirtWorker', ClassWithInterface),
        SimpleFakeEntry('OpenStackLatentWorker', ClassWithInterface),
        SimpleFakeEntry('newthirdparty', ClassWithInterface),
        SimpleFakeEntry('deep.newthirdparty', ClassWithInterface),
    ],
    'buildbot.util': [
        SimpleFakeEntry('WorkerLock', ClassWithInterface),
        SimpleFakeEntry('enforceChosenWorker', ClassWithInterface),
        SimpleFakeEntry('WorkerChoiceParameter', ClassWithInterface),
    ],
}


def provide_worker_fake_entries(group):
    """
    give a set of fake entries for known groups
    """
    return _WORKER_FAKE_ENTRIES.get(group, [])
